import json
import re

from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union
from urllib.parse import quote_plus

from .helpers import Actions, BaseError, Context, Cardinality, LastModificationDate, WithId
from .connection import Connection
from .model import Model
from .langstring import Languages, LangStringParam, LangString
from .listnode import ListNode

class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


@strict
class PropertyClass(Model):
    _context: Context
    _id: str
    _name: str
    _ontology_id: str
    _superproperties: List[str]
    _object: str
    _subject: str
    _gui_element: str
    _gui_attributes: Dict[str, str]
    _label: LangString
    _comment: LangString
    _editable: bool
    _linkvalue: bool

    def __init__(self,
                 con: Connection,
                 context: Context,
                 id: Optional[str] = None,
                 name: Optional[str] = None,
                 ontology_id: Optional[str] = None,
                 superproperties: Optional[List[Union['PropertyClass', str]]] = None,
                 object: Optional[str] = None,
                 subject: Optional[str] = None,
                 gui_element: Optional[str] = None,
                 gui_attributes: Optional[Dict[str, str]] = None,
                 label: Optional[Union[LangString, str]] = None,
                 comment: Optional[Union[LangString, str]] = None,
                 editable: Optional[bool] = None,
                 linkvalue: Optional[bool] = None):
        super().__init__(con)
        if not isinstance(context, Context):
            raise BaseError('"context"-parameter must be an instance of Context')
        self._context = context
        self._id = id
        self._name = name
        self._ontology_id = ontology_id
        if isinstance(superproperties, PropertyClass):
            self._superproperties = list(map(lambda a: a.id, superproperties))
        else:
            self._superproperties = superproperties
        self._object = object
        self._subject = subject
        self._gui_element = gui_element
        self._gui_attributes = gui_attributes
        #
        # process label
        #
        if label is not None:
            if isinstance(label, str):
                self._label = LangString(label)
            elif isinstance(label, LangString):
                self._label = label
            else:
                raise BaseError('Invalid LangString for label!')
        else:
            self._label = None
        #
        # process comment
        #
        if comment is not None:
            if isinstance(comment, str):
                self._comment = LangString(comment)
            elif isinstance(comment, LangString):
                self._comment = comment
            else:
                raise BaseError('Invalid LangString for comment!')
        else:
            self._comment = None

        self._editable = editable
        self._linkvalue = linkvalue

    #
    # Here follows a list of getters/setters
    #
    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        raise BaseError('"name" cannot be modified!')

    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        raise BaseError('"id" cannot be modified!')

    @property
    def ontology_id(self) -> Optional[str]:
        return self._ontology_id

    @ontology_id.setter
    def ontology_id(self, value: str) -> None:
        raise BaseError('"ontology_id" cannot be modified!')

    @property
    def superproperties(self) -> Optional[List[str]]:
        return self._superproperties

    @superproperties.setter
    def superproperties(self, value: str) -> None:
        raise BaseError('"superproperties" cannot be modified!')

    @property
    def object(self) -> Optional[str]:
        return self._object

    @object.setter
    def object(self, value: Any):
        raise BaseError('"object" cannot be modified!')

    @property
    def subject(self) -> Optional[str]:
        return self._subject

    @subject.setter
    def subject(self, value: Any):
        raise BaseError('"subject" cannot be modified!')

    @property
    def gui_element(self) -> Optional[str]:
        return self._gui_element

    @gui_element.setter
    def gui_element(self, value: str) -> None:
        self._gui_element = value
        self._changed.append('gui_element')

    @property
    def gui_attributes(self) -> Optional[Dict[str, str]]:
        return self._gui_attributes

    @gui_attributes.setter
    def gui_attributes(self, value: List[Dict[str, str]]) -> None:
        self._gui_attributes = value
        self._changed.append('gui_attributes')

    def addGuiAttribute(self, key: str, value: str) -> None:
        self._gui_attributes[key] = value
        self._changed.append('gui_attributes')

    def rmGuiAttribute(self, key: str) -> None:
        if self._gui_attributes.get(key) is not None:
            del self._gui_attributes[key]
            self._changed.append('gui_attributes')

    @property
    def label(self) -> Optional[LangString]:
        return self._label

    @label.setter
    def label(self, value: Optional[Union[LangString, str]]) -> None:
        if value is None:
            self._label.empty()  # clear all labels
        else:
            if isinstance(value, LangString):
                self._label = value
            elif isinstance(value, str):
                self._label = LangString(value)
            else:
                raise BaseError('Not a valid LangString')
        self._changed.add('label')

    def addLabel(self, lang: Union[Languages, str], value: str) -> None:
        self._label[lang] = value
        self._changed.add('label')

    def rmLabel(self, lang: Union[Languages, str]) -> None:
        del self._label[lang]
        self._changed.add('label')

    @property
    def comment(self) -> Optional[LangString]:
        return self._comment

    @comment.setter
    def comment(self, value: Optional[LangString]) -> None:
        if value is None:
            self._comment.empty()  # clear all comments!
        else:
            if isinstance(value, LangString):
                self._comment = value
            elif isinstance(value, str):
                self._comment = LangString(value)
            else:
                raise BaseError('Not a valid LangString')
        self._changed.add('comment')

    def addComment(self, lang: Union[Languages, str], value: str) -> None:
        self._comment[lang] = value
        self._changed.add('comment')

    def rmComment(self, lang: Union[Languages, str]) -> None:
        del self._comment[lang]
        self._changed.add('comment')

    @property
    def editable(self) -> bool:
        return self._editable

    @editable.setter
    def editable(self, value: bool) -> None:
        raise BaseError('"editable" cannot be modified!')

    @property
    def linkvalue(self) -> bool:
        return self._linkvalue

    @linkvalue.setter
    def linkvalue(self) -> None:
        raise BaseError('"linkvalue" cannot be modified!')

    @classmethod
    def fromJsonObj(cls, con: Connection, context: Context, json_obj: Any) -> Any:
        if isinstance(json_obj, list):
            json_obj = json_obj[0]  # TODO: Is it possible to have more than one element in the list??
        if not isinstance(con, Connection):
            raise BaseError('"con"-parameter must be an instance of Connection')
        if not isinstance(context, Context):
            raise BaseError('"context"-parameter must be an instance of Context')
        rdf = context.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        xsd = context.prefix_from_iri("http://www.w3.org/2001/XMLSchema#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui = context.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")

        if not (json_obj.get(knora_api + ':isResourceProperty')):
            raise BaseError("This is not a property!")
        if json_obj.get('@id') is None:
            raise BaseError('Property class has no "@id"!')
        tmp_id = json_obj.get('@id').split(':')
        id = context.iri_from_prefix(tmp_id[0]) + '#' + tmp_id[1]
        ontology_id = tmp_id[0]
        name = tmp_id[1]
        superproperties_obj = json_obj.get(rdfs + ':subPropertyOf')
        superproperties: List[Union[None, str]]
        if not isinstance(superproperties_obj, list):
            superproperties_obj = [superproperties_obj]  # make a list out of it
        if superproperties_obj is not None:
            superprops: List[Any] = list(filter(lambda a: a.get('@id') is not None, superproperties_obj))
            superproperties = list(map(lambda a: a['@id'], superprops))
        else:
            superproperties = None
        object = WithId(json_obj.get(knora_api + ':objectType')).str()
        subject = WithId(json_obj.get(knora_api + ':subjectType')).str()
        label = LangString.fromJsonLdObj(json_obj.get(rdfs + ':label'))
        comment = LangString.fromJsonLdObj(json_obj.get(rdfs + ':comment'))
        gui_element = None
        if json_obj.get(salsah_gui + ':guiElement') is not None:
            gui_element = WithId(json_obj.get(salsah_gui + ':guiElement')).str()
        gui_attributes_list = json_obj.get(salsah_gui + ':guiAttribute')
        gui_attributes: Union[None, Dict[str, str]] = None
        if gui_attributes_list is not None:
            gui_attributes = {}
            if not isinstance(gui_attributes_list, list):
                gui_attributes_list = [gui_attributes_list]
            for ga in gui_attributes_list:
                tmp = ga.split('=')
                if len(tmp) == 1:
                    gui_attributes[tmp[0]] = ''
                else:
                    gui_attributes[tmp[0]] = tmp[1]

        editable = json_obj.get(knora_api + ':isEditable')
        linkvalue = json_obj.get(knora_api + ':isLinkProperty')
        return cls(con=con,
                   context=context,
                   id=id,
                   name=name,
                   ontology_id=ontology_id,
                   superproperties=superproperties,
                   object=object,
                   subject=subject,
                   gui_element=gui_element,
                   gui_attributes=gui_attributes,
                   label=label,
                   comment=comment,
                   editable=editable,
                   linkvalue=linkvalue)

    def toJsonObj(self, lastModificationDate: LastModificationDate, action: Actions, what: Optional[str] = None) -> Any:

        def resolve_propref(resref: str):
            tmp = resref.split(':')
            if len(tmp) > 1:
                if tmp[0]:
                    return {"@id": resref}  # fully qualified name in the form "prefix:name"
                else:
                    return {"@id": self._context.prefix_from_iri(self._ontology_id) + ':' + tmp[1]}  # ":name" in current ontology
            else:
                return {"@id": "knora-api:" + resref}  # no ":", must be from knora-api!

        tmp = {}
        exp = re.compile('^http.*')  # It is already a fully IRI
        if exp.match(self._ontology_id):
            propid = self._context.prefix_from_iri(self._ontology_id) + ":" + self._name
            ontid = self._ontology_id
        else:
            propid = self._ontology_id + ":" + self._name
            ontid = self._context.iri_from_prefix(self._ontology_id)
        if action == Actions.Create:
            if self._name is None:
                raise BaseError("There must be a valid property class name!")
            if self._ontology_id is None:
                raise BaseError("There must be a valid ontology_id given!")
            if self._superproperties is None:
                superproperties = [{"@id": "knora-api:hasValue"}]
            else:
                superproperties = list(map(resolve_propref, self._superproperties))

            tmp = {
                "@id": ontid,  # self._ontology_id,
                "@type": "owl:Ontology",
                "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
                "@graph": [{
                    "@id": propid,
                    "@type": "owl:ObjectProperty",
                    "rdfs:label": self._label.toJsonLdObj(),
                    "rdfs:subPropertyOf": superproperties
                }],
                "@context": self._context.toJsonObj()
            }
            if self._comment is not None:
                if not self._comment.isEmpty():
                    tmp['@graph'][0]["rdfs:comment"] = self._comment.toJsonLdObj()
            if self._subject is not None:
                tmp['@graph'][0]["knora-api:subjectType"] = resolve_propref(self._subject)
            if self._object is not None:
                tmp['@graph'][0]["knora-api:objectType"] = resolve_propref(self._object)
            if self._gui_element is not None:
                tmp['@graph'][0]["salsah-gui:guiElement"] = {
                    "@id": self._gui_element
                }
            if self._gui_attributes:
                ga = list(map(lambda x: x[0] + '=' + str(x[1]), self._gui_attributes.items()))
                tmp['@graph'][0]["salsah-gui:guiAttribute"] = ga
        elif action == Actions.Update:
            tmp = {
                "@id": ontid,  # self._ontology_id,
                "@type": "owl:Ontology",
                "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
                "@graph": [{
                    "@id": propid,
                    "@type": "owl:ObjectProperty",
                }],
                "@context": self._context.toJsonObj(),
            }
            if what == 'label':
                if not self._label.isEmpty() and 'label' in self._changed:
                    tmp['@graph'][0]['rdfs:label'] = self._label.toJsonLdObj()
            if what == 'comment':
                if not self._comment.isEmpty() and 'comment' in self._changed:
                    tmp['@graph'][0]['rdfs:comment'] = self._comment.toJsonLdObj()

        return tmp

    def create(self, last_modification_date: LastModificationDate) -> Tuple[LastModificationDate, 'PropertyClass']:
        jsonobj = self.toJsonObj(last_modification_date, Actions.Create)
        jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=2)
        result = self._con.post('/v2/ontologies/properties', jsondata)
        last_modification_date = LastModificationDate(result['knora-api:lastModificationDate'])
        return last_modification_date, PropertyClass.fromJsonObj(self._con, self._context, result['@graph'])

    def update(self, last_modification_date: LastModificationDate) -> Tuple[LastModificationDate, 'ResourceClass']:
        #
        # Note: Knora is able to change only one thing per call, either label or comment!
        #
        result = None
        something_changed = False
        if 'label' in self._changed:
            jsonobj = self.toJsonObj(last_modification_date, Actions.Update, 'label')
            jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=4)
            result = self._con.put('/v2/ontologies/properties', jsondata)
            last_modification_date = LastModificationDate(result['knora-api:lastModificationDate'])
            something_changed = True
        if 'comment' in self._changed:
            jsonobj = self.toJsonObj(last_modification_date, Actions.Update, 'comment')
            jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=4)
            result = self._con.put('/v2/ontologies/properties', jsondata)
            last_modification_date = LastModificationDate(result['knora-api:lastModificationDate'])
            something_changed = True
        if something_changed:
            return last_modification_date, PropertyClass.fromJsonObj(self._con, self._context, result['@graph'])
        else:
            return last_modification_date, self

    def delete(self, last_modification_date: LastModificationDate) -> LastModificationDate:
        result = self._con.delete('/v2/ontologies/properties/' + quote_plus(self._id) + '?lastModificationDate=' + str(last_modification_date))
        return LastModificationDate(result['knora-api:lastModificationDate'])

    def createDefinitionFileObj(self, context: Context, shortname: str):
        """
        Create an object that jsonfied can be used as input to "create_onto"

        :param context: Context of the ontology
        :param shortname: Shortname of the ontology
        :return: Python object to be jsonfied
        """
        property = {
            "name": self._name
        }
        if self._object is not None:
            property["name"] = self._name
        if self._superproperties is not None:
            superprops = []
            for sc in self._superproperties:
                superprops.append(context.reduce_iri(sc, shortname))
            property["super"] = superprops
        if self._object is not None:
            property["object"] = context.reduce_iri(self._object, shortname)
        if self._label is not None:
            property["labels"] = self._label.createDefinitionFileObj()
        if self._gui_element is not None:
            property["gui_element"] = context.reduce_iri(self._gui_element, shortname)
        if self._gui_attributes:
            gui_elements = {}
            for (attname, attvalue) in self._gui_attributes.items():
                if attname == "size":
                    gui_elements[attname] = int(attvalue)
                elif attname == "maxsize":
                    gui_elements[attname] = int(attvalue)
                elif attname == "hlist":
                    iri = attvalue[1:-1]
                    rootnode = ListNode(con=self._con, id=iri).read()
                    gui_elements[attname] = rootnode.name
                elif attname == "numprops":
                    gui_elements[attname] = int(attvalue)
                elif attname == "ncolors":
                    gui_elements[attname] = int(attvalue)
                elif attname == "cols":
                    gui_elements[attname] = int(attvalue)
                elif attname == "rows":
                    gui_elements[attname] = int(attvalue)
                elif attname == "width":
                    gui_elements[attname] = str(attvalue)
                elif attname == "wrap":
                    gui_elements[attname] = str(attvalue)
                elif attname == "max":
                    gui_elements[attname] = float(attvalue)
                elif attname == "min":
                    gui_elements[attname] = float(attvalue)
                else:
                    gui_elements[attname] = str(attvalue)
            property["gui_attributes"] = gui_elements
        return property

    def print(self, offset: int = 0):
        blank = ' '
        print(f'{blank:>{offset}}Property Class Info')
        print(f'{blank:>{offset+2}}Name:            {self._name}')
        print(f'{blank:>{offset+2}}Ontology prefix: {self._ontology_id}')
        print(f'{blank:>{offset+2}}Superproperties:')
        if self._superproperties is not None:
            for sc in self._superproperties:
                print(f'{blank:>{offset + 4}}{sc}')
        if self._label is not None:
            print(f'{blank:>{offset + 2}}Labels:')
            self._label.print(offset + 4)
        if self._subject is not None:
            print(f'{blank:>{offset + 4}}Subject: {self._subject}')
        if self._object is not None:
            print(f'{blank:>{offset + 4}}Object: {self._object}')
        if self._gui_element is not None:
            print(f'{blank:>{offset + 4}}Guielement: {self._gui_element}')
        if self._gui_attributes is not None:
            print(f'{blank:>{offset + 4}}Gui Attributes:')
            if self._gui_attributes is not None:
                for (attname, attvalue) in self._gui_attributes.items():
                    print(f'{blank:>{offset + 6}}Attribute: {attname} Value: {attvalue}')
        if self._comment is not None:
            print(f'{blank:>{offset + 2}}Comments:')
            self._comment.print(offset + 4)
        if self._editable is not None:
            print(f'{blank:>{offset + 4}}Editable: {self._editable}')
        if self._linkvalue is not None:
            print(f'{blank:>{offset + 4}}Editable: {self._linkvalue}')





import unittest

from dsplib.models.connection import Connection
from dsplib.models.langstring import Languages, LangStringParam, LangString
from dsplib.models.project import Project


class TestProject(unittest.TestCase):

    def test_project(self):
        con = Connection('http://0.0.0.0:3333')
        project = Project(
            con=con,
            id='http://rdfh.ch/gaga',
            shortcode='0FF0',
            shortname="pytest",
            longname="Python testing",
            description=LangString({Languages.EN: 'knora-py testing', Languages.DE: 'Tests für knora-py'}),
            keywords={'test', 'testing', 'gaga'},
            selfjoin=False,
            status=True,
            logo='gaga.gif'
        )
        self.assertIsNotNone(project)
        self.assertEqual(project.id, 'http://rdfh.ch/gaga')
        self.assertEqual(project.shortcode, '0FF0')
        self.assertEqual(project.shortname, 'pytest')
        self.assertEqual(project.longname, 'Python testing')
        self.assertEqual(project.description['en'], 'knora-py testing')
        self.assertEqual(project.description['de'], 'Tests für knora-py')
        self.assertEqual(project.selfjoin, False)
        self.assertEqual(project.status, True)
        self.assertEqual(project.keywords, {'test', 'testing', 'gaga'})
        self.assertEqual(project.logo, 'gaga.gif')

    def test_read(self):
        con = Connection('http://0.0.0.0:3333')
        project = Project(
            con=con,
            id='http://rdfh.ch/projects/0001'
        ).read()
        self.assertIsNotNone(project)
        self.assertEqual(project.id, 'http://rdfh.ch/projects/0001')
        self.assertEqual(project.shortcode, '0001')
        self.assertEqual(project.shortname, 'anything')
        self.assertEqual(project.longname, 'Anything Project')
        self.assertEqual(project.description[None], 'Anything Project')
        self.assertEqual(project.selfjoin, False)
        self.assertEqual(project.status, True)

    def test_create(self):
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        project = Project(
            con=con,
            shortcode='0FF0',
            shortname="pytest0",
            longname="Python testing 0",
            description=LangString({Languages.EN: 'knora-py testing', Languages.DE: 'Tests für knora-py'}),
            keywords={'test', 'testing', 'gaga'},
            selfjoin=False,
            status=True,
            logo='gaga.gif'
        )
        nproject = project.create()
        self.assertIsNotNone(nproject)
        self.assertEqual(nproject.shortcode, '0FF0')
        self.assertEqual(nproject.shortname, 'pytest0')
        self.assertEqual(nproject.longname, 'Python testing 0')
        self.assertEqual(nproject.description['en'], 'knora-py testing')
        self.assertEqual(nproject.description['de'], 'Tests für knora-py')
        self.assertEqual(nproject.selfjoin, False)
        self.assertEqual(nproject.status, True)
        self.assertEqual(nproject.keywords, {'test', 'testing', 'gaga'})

    def test_update(self):
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        project = Project(
            con=con,
            shortcode='0FF1',
            shortname="pytest1",
            longname="Python testing 1",
            description=LangString({Languages.EN: 'knora-py testing', Languages.DE: 'Tests für knora-py'}),
            keywords={'test', 'testing', 'gaga'},
            selfjoin=False,
            status=True,
            logo='gaga.gif'
        ).create()

        project.shortname = "pytest1a"
        project.longname = "Python testing 1a"
        project.addDescription('fr', 'Les testes pour python')
        project.rmDescription('de')
        project.selfjoin = True
        project.status = False
        project.rmKeyword('gaga')
        project.addKeyword('Guguseli')
        nproject = project.update()

        self.assertIsNotNone(nproject)
        self.assertEqual(nproject.shortcode, '0FF1')
        # self.assertEqual(nnproject.shortname, 'pytest2') # Does not work....? Knora-API problem
        self.assertEqual(nproject.longname, 'Python testing 1a')
        self.assertEqual(nproject.description['en'], 'knora-py testing')
        self.assertEqual(nproject.description['fr'], 'Les testes pour python')
        self.assertEqual(nproject.selfjoin, True)
        self.assertEqual(nproject.status, False)
        self.assertEqual(nproject.keywords, {'test', 'testing', 'Guguseli'})

    def test_delete(self):
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        project = Project(
            con=con,
            shortcode='0FF2',
            shortname="pytest2",
            longname="Python testing 2",
            description=LangString({Languages.EN: 'knora-py testing', Languages.DE: 'Tests für knora-py'}),
            keywords={'test', 'testing', 'gaga'},
            selfjoin=False,
            status=True,
            logo='gaga.gif'
        ).create()

        nproject = project.delete()
        self.assertIsNotNone(nproject)
        self.assertEqual(nproject.shortcode, '0FF2')
        self.assertEqual(nproject.shortname, 'pytest2')
        self.assertEqual(nproject.longname, 'Python testing 2')
        self.assertEqual(nproject.description['en'], 'knora-py testing')
        self.assertEqual(nproject.description['de'], 'Tests für knora-py')
        self.assertEqual(nproject.selfjoin, False)
        self.assertEqual(nproject.status, False)
        self.assertEqual(nproject.keywords, {'test', 'testing', 'gaga'})

if __name__ == '__main__':
    unittest.main()



import unittest

from dsplib.models.connection import Connection
from dsplib.models.langstring import Languages, LangStringParam, LangString
from dsplib.models.user import User


class TestUser(unittest.TestCase):

    def setUp(self) -> None:
        self.con = Connection('http://0.0.0.0:3333')
        self.con.login('root@example.com', 'test')

    def test_user(self):
        user = User(
            con=self.con,
            username='wilee',
            email='wilee.coyote@canyon.com',
            givenName='Wile E.',
            familyName='Coyote',
            password='BeepBeep',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        )
        self.assertEqual(user.username, 'wilee')
        self.assertEqual(user.email, 'wilee.coyote@canyon.com')
        self.assertEqual(user.givenName, 'Wile E.')
        self.assertEqual(user.familyName, 'Coyote')
        self.assertTrue(user.status)
        self.assertEqual(user.lang, Languages.EN)
        self.assertTrue(user.sysadmin)
        self.assertEqual(user.in_projects, {"http://rdfh.ch/projects/0001": True})
        self.assertEqual(user.in_groups, {"http://rdfh.ch/groups/0001/thing-searcher"})

    def test_User_read(self):
        user = User(
            con=self.con,
            id='http://rdfh.ch/users/91e19f1e01'
        ).read()
        self.assertEqual(user.id, 'http://rdfh.ch/users/91e19f1e01')
        self.assertEqual(user.username, 'root-alt')
        self.assertEqual(user.familyName, 'Admin-alt')
        self.assertEqual(user.givenName, 'Administrator-alt')
        self.assertEqual(user.lang, Languages.DE)
        self.assertTrue(user.status)
        self.assertFalse(user.sysadmin)
        self.assertEqual(user.in_projects, {"http://rdfh.ch/projects/0803": False})

    def test_User_create(self):
        user = User(
            con=self.con,
            username='wilee',
            email='wilee.coyote@canyon.com',
            givenName='Wile E.',
            familyName='Coyote',
            password='BeepBeep',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        self.assertEqual(user.username, 'wilee')
        self.assertEqual(user.email, 'wilee.coyote@canyon.com')
        self.assertEqual(user.givenName, 'Wile E.')
        self.assertEqual(user.familyName, 'Coyote')
        self.assertTrue(user.status)
        self.assertEqual(user.lang, Languages.EN)
        self.assertTrue(user.sysadmin)
        self.assertEqual(user.in_projects, {"http://rdfh.ch/projects/0001": True})
        self.assertEqual(user.in_groups, {"http://rdfh.ch/groups/0001/thing-searcher"})

    def test_User_read2(self):
        user = User(
            con=self.con,
            username='wilee2',
            email='wilee.coyote2@canyon.com',
            givenName='Wile E.2',
            familyName='Coyote2',
            password='BeepBeep2',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        user = User(
            con=self.con,
            email='wilee.coyote2@canyon.com'
        ).read()
        self.assertEqual(user.username, 'wilee2')
        self.assertEqual(user.familyName, 'Coyote2')
        self.assertEqual(user.givenName, 'Wile E.2')
        self.assertEqual(user.lang, Languages.EN)
        self.assertTrue(user.status)
        self.assertTrue(user.sysadmin)
        self.assertEqual(user.in_projects, {"http://rdfh.ch/projects/0001": True})
        self.assertEqual(user.in_groups, {"http://rdfh.ch/groups/0001/thing-searcher"})

    def test_User_delete(self):
        user = User(
            con=self.con,
            username='wilee3',
            email='wilee.coyote3@canyon.com',
            givenName='Wile E.3',
            familyName='Coyote3',
            password='BeepBeep3',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        nuser = user.delete()
        self.assertIsNotNone(nuser)
        self.assertFalse(nuser.status)

    def test_User_update(self):
        user = User(
            con=self.con,
            username='wilee4',
            email='wilee.coyote4@canyon.com',
            givenName='Wile E.4',
            familyName='Coyote4',
            password='BeepBeep4',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        user.email = 'roadrunner.geococcyx@canyon.com'
        user.username = 'roadrunner'
        user.givenName = 'roadrunner'
        user.familyName = 'Geococcyx'
        user.lang = 'fr'
        user.status = False
        user.sysadmin = False
        nuser = user.update()
        self.assertEqual(nuser.email, 'roadrunner.geococcyx@canyon.com')
        self.assertEqual(nuser.username, 'roadrunner')
        self.assertEqual(nuser.givenName, 'roadrunner')
        self.assertEqual(nuser.familyName, 'Geococcyx')
        self.assertEqual(nuser.lang, Languages.FR)
        self.assertFalse(nuser.status)
        self.assertFalse(nuser.sysadmin)

    def test_User_update2(self):
        user = User(
            con=self.con,
            username='wilee5',
            email='wilee.coyote5@canyon.com',
            givenName='Wile E.5',
            familyName='Coyote5',
            password='BeepBeep5',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        user.password = 'gagagagagagagaga'
        nuser = user.update('test')
        self.assertIsNotNone(nuser)

    def test_User_addToGroup(self):
        return #  TODO: Check why this test failes with error: Message:{"error":"org.knora.webapi.UpdateNotPerformedException: User's 'group' memberships where not updated. Please report this as a possible bug."}
        user = self.createTestUser()
        user.addToGroup('http://rdfh.ch/groups/0001/thing-searcher')
        nuser = user.update()
        self.assertIsNotNone(nuser)
        self.assertEqual(nuser.in_groups, {"http://rdfh.ch/groups/0001/thing-searcher", 'http://rdfh.ch/groups/0001/thing-searcher'})

    def test_User_rmFromGroup(self):
        user = User(
            con=self.con,
            username='wilee6',
            email='wilee.coyote6@canyon.com',
            givenName='Wile E.6',
            familyName='Coyote6',
            password='BeepBeep6',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/00FF/images-reviewer"}
        ).create()
        user.rmFromGroup('http://rdfh.ch/groups/00FF/images-reviewer')
        nuser = user.update()
        self.assertIsNotNone(nuser)

    def test_User_addToProject(self):
        user = User(
            con=self.con,
            username='wilee7',
            email='wilee.coyote7@canyon.com',
            givenName='Wile E.7',
            familyName='Coyote7',
            password='BeepBeep7',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        user.addToProject('http://rdfh.ch/projects/00FF', False)
        nuser = user.update()
        self.assertIsNotNone(nuser)
        self.assertEqual(nuser.in_projects, {"http://rdfh.ch/projects/0001": True, 'http://rdfh.ch/projects/00FF': False})

    def test_User_rmFromProject(self):
        user = User(
            con=self.con,
            username='wilee8',
            email='wilee.coyote8@canyon.com',
            givenName='Wile E.8',
            familyName='Coyote8',
            password='BeepBeep8',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        user.rmFromProject('http://rdfh.ch/projects/0001')
        nuser = user.update()
        self.assertIsNotNone(nuser)
        self.assertEqual(nuser.in_projects, {})

    def test_User_unmakeProjectAdmin(self):
        user = User(
            con=self.con,
            username='wilee9',
            email='wilee.coyote9@canyon.com',
            givenName='Wile E.9',
            familyName='Coyote9',
            password='BeepBeep9',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        user.unmakeProjectAdmin('http://rdfh.ch/projects/0001')
        nuser = user.update()
        self.assertIsNotNone(nuser)
        self.assertEqual(nuser.in_projects, {'http://rdfh.ch/projects/0001': False})

    def test_User_makeProjectAdmin(self):
        user = User(
            con=self.con,
            username='wileeA',
            email='wilee.coyoteA@canyon.com',
            givenName='Wile E.A',
            familyName='CoyoteA',
            password='BeepBeepA',
            lang=Languages.EN,
            status=True,
            sysadmin=True,
            in_projects={"http://rdfh.ch/projects/0001": True},
            in_groups={"http://rdfh.ch/groups/0001/thing-searcher"}
        ).create()
        user.addToProject('http://rdfh.ch/projects/00FF', False)
        user = user.update()
        user.makeProjectAdmin('http://rdfh.ch/projects/00FF')
        nuser = user.update()
        self.assertIsNotNone(nuser)
        self.assertEqual(nuser.in_projects, {'http://rdfh.ch/projects/0001': True,
                                             'http://rdfh.ch/projects/00FF': True})

    def test_getAllUsers(self):
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')
        all_users = User.getAllUsers(con)
        for u in all_users:
            self.assertIsNotNone(u.id)

if __name__ == '__main__':
    unittest.main()



import requests
from .helpers import BaseError

import os

class Sipi:


    def __init__(self, sipiserver: str, token: str):
        self.sipiserver = sipiserver
        self.token = token

    def on_api_error(self, res):
        """
        Method to check for any API errors
        :param res: The input to check, usually JSON format
        :return: Possible KnoraError that is being raised
        """

        if res.status_code != 200:
            raise BaseError("SIPI-ERROR: status code=" + str(res.status_code) + "\nMessage:" + res.text)

        if 'error' in res:
            raise BaseError("SIPI-ERROR: API error: " + res.error)

    def upload_bitstream(self, filepath):
        print(f"filepath=${os.path.basename(filepath)} (${filepath})")
        files = {
            'file': (os.path.basename(filepath), open(filepath, 'rb')),
        }
        req = requests.post(self.sipiserver + "/upload?token=" + self.token,
                            files=files)
        self.on_api_error(req)
        res = req.json()
        return res



import os
import json
from jsonschema import validate

from ..models.helpers import Actions, BaseError, Context, Cardinality
from ..models.connection import Connection
from ..models.project import Project
from ..models.listnode import ListNode
from .onto_commons import list_creator, validate_list_from_excel, json_list_from_excel

def create_lists (input_file: str, output_file: str, server: str, user: str, password: str, verbose: bool) -> bool:
    current_dir = os.path.dirname(os.path.realpath(__file__))

    # let's read the schema for the data model definition
    with open(os.path.join(current_dir, 'knora-schema-lists.json')) as s:
        schema = json.load(s)
    # read the data model definition
    with open(input_file) as f:
        datamodel = json.load(f)

    #
    # now let's see if there are any lists defined as reference to excel files
    #
    lists = datamodel["project"].get('lists')
    if lists is not None:
        newlists: [] = []
        for rootnode in lists:
            if rootnode.get("nodes") is not None and isinstance(rootnode["nodes"], dict) and rootnode["nodes"].get("file") is not None:
                newroot = {
                    "name": rootnode.get("name"),
                    "labels": rootnode.get("labels"),
                    "comments": rootnode.get("comments")
                }
                startrow = 1 if rootnode["nodes"].get("startrow") is None else rootnode["nodes"]["startrow"]
                startcol = 1 if rootnode["nodes"].get("startcol") is None else rootnode["nodes"]["startcol"]
                json_list_from_excel(rootnode=newroot,
                                     filepath=rootnode["nodes"]["file"],
                                     sheetname=rootnode["nodes"]["worksheet"],
                                     startrow=startrow,
                                     startcol=startcol)
                newlists.append(newroot)
            else:
                newlists.append(rootnode)
        datamodel["project"]["lists"] = newlists

    # validate the data model definition in order to be sure that it is correct
    validate(datamodel, schema)

    if verbose:
        print("Data model is syntactically correct and passed validation!")

    #
    # Connect to the DaSCH Service Platform API
    #
    con = Connection(server)
    con.login(user, password)

    # --------------------------------------------------------------------------
    # let's read the prefixes of external ontologies that may be used
    #
    context = Context(datamodel["prefixes"])

    # --------------------------------------------------------------------------
    # Let's get the project which must exist
    #
    project = Project(
        con=con,
        shortcode=datamodel["project"]["shortcode"],
    ).read()
    assert project is not None

    # --------------------------------------------------------------------------
    # now let's create the lists
    #
    if verbose:
        print("Creating lists...")
    lists = datamodel["project"].get('lists')
    listrootnodes = {}
    if lists is not None:
        for rootnode in lists:
            if verbose is not None:
                print("  Creating list:" + rootnode['name'])
            root_list_node = ListNode(
                con=con,
                project=project,
                label=rootnode['labels'],
                comment=rootnode.get('comments'),
                name=rootnode['name']
            ).create()
            listnodes = list_creator(con, project, root_list_node, rootnode['nodes'])
            listrootnodes[rootnode['name']] = {
                "id": root_list_node.id,
                "nodes": listnodes
            }

    with open(output_file, 'w', encoding="utf-8") as fp:
        json.dump(listrootnodes, fp, indent=3, sort_keys=True)
        print("The definitions of the node-id's can be found in \"{}}\"!".format(output_file))
    return True



from dataclasses import dataclass
from typing import NewType, List, Set, Dict, Tuple, Optional, Any, Union, Pattern
from enum import Enum, unique
from traceback import format_exc
from pystrict import strict
import re
import sys

#
# here we do some data typing that should help
#
@dataclass
class OntoInfo:
    """
    A small class thats holds an ontology IRI. The variable "hashtag" is True, if "#" is used s separate elements,
    False if the element name is just appended
    """
    iri: str
    hashtag: bool


ContextType = NewType("ContextType", Dict[str, OntoInfo])


def LINE():
    return sys._getframe(1).f_lineno


@strict
class IriTest:
    __iri_regexp = re.compile("^(http)s?://([\\w\\.\\-~]+)?(:\\d{,6})?(/[\\w\\-~]+)*(#[\\w\\-~]*)?")

    @classmethod
    def test(cls, val: str) -> bool:
        m = cls.__iri_regexp.match(val)
        return m.span()[1] == len(val) if m else False


@strict
class BaseError(Exception):
    """
    A basic error class
    """
    _message: str

    def __init__(self, message: str) -> None:
        """
        Constructor for error message
        :param message: error message string
        """
        super().__init__()
        self._message = message

    def __str__(self) -> str:
        """
        Convert to string
        :return: stringyfied error message
        """
        return "ERROR: " + self._message + "!\n\n" + format_exc()

    @property
    def message(self) -> str:
        return self._message


@unique
class Actions(Enum):
    Create = 1
    Read = 2
    Update = 3
    Delete = 4


@unique
class Cardinality(Enum):
    C_1 = "1"
    C_0_1 = "0-1"
    C_1_n = "1-n"
    C_0_n = "0-n"

@strict
class ContextIterator:
    _context: 'Context'
    _prefixes: List[str]
    _index: int

    def __init__(self, context: 'Context'):
        self._context = context
        self._prefixes = [x for x in self._context.context]
        self._index = 0

    def __next__(self):
        if len(self._context.context) == 0 and self._index == 0:
            return None, None
        elif self._index < len(self._context.context):
            tmp = self._prefixes[self._index]
            self._index += 1
            return tmp, self._context.context[tmp]
        else:
            raise StopIteration


@strict
class Context:
    """
    This class holds a JSON-LD context with the ontology IRI's and the associated prefixes
    """
    _context: ContextType
    _rcontext: Dict[str, str]
    _exp: Pattern

    common_ontologies: ContextType = {
        "foaf": OntoInfo("http://xmlns.com/foaf/0.1/", False),
        "dc": OntoInfo("http://purl.org/dc/elements/1.1/", False),
        "dcterms": OntoInfo("http://purl.org/dc/terms/", False),
        "dcmi": OntoInfo("http://purl.org/dc/dcmitype/", False),
        "skos": OntoInfo("http://www.w3.org/2004/02/skos/core", True),
        "bibtex": OntoInfo("http://purl.org/net/nknouf/ns/bibtex", True),
        "bibo": OntoInfo("http://purl.org/ontology/bibo/", False),
        "cidoc": OntoInfo("http://purl.org/NET/cidoc-crm/core", True)
    }

    knora_ontologies: ContextType = {
        "knora-api": OntoInfo("http://api.knora.org/ontology/knora-api/v2", True),
        "salsah-gui": OntoInfo("http://api.knora.org/ontology/salsah-gui/v2", True)
    }

    base_ontologies: ContextType = {
        "rdf": OntoInfo("http://www.w3.org/1999/02/22-rdf-syntax-ns", True),
        "rdfs": OntoInfo("http://www.w3.org/2000/01/rdf-schema", True),
        "owl": OntoInfo("http://www.w3.org/2002/07/owl", True),
        "xsd": OntoInfo("http://www.w3.org/2001/XMLSchema", True)
    }

    def __is_iri(self, val: str) -> bool:
        """
        Test if string has a valid IRI pattern

        :param val: Input string
        :return: True, if val corresponds to IRI pattern
        """

        m = self._exp.match(val)
        return m.span()[1] == len(val) if m else False

    def __init__(self, context: Optional[Dict[str, str]] = None):
        """
        THe Constructor of the Context. It takes one optional parameter which as a dict of
        prefix - ontology-iri pairs. If the hashtag "#" is used to append element name, the
        ontology-iri *must* end with "#"!
        :param context: A dict of prefix - ontology-iri pairs
        """
        #
        # regexp to test for a complete IRI (including fragment identifier)
        #
        self._exp = re.compile("^(http)s?://([\\w\\.\\-~]+)?(:\\d{,6})?(/[\\w\\-~]+)*(#[\\w\\-~]*)?")

        if context:
            cleaned_input: Dict[str, str] = {prefix: onto for (prefix, onto) in context.items() if self.base_ontologies.get(prefix) is None and self.knora_ontologies.get(prefix) is None}
            self._context = {}
            for prefix, onto in cleaned_input.items():
                self._context[prefix] = OntoInfo(onto[:-1], True) if onto.endswith('#') else OntoInfo(onto, False)
            #
            # we always want the base ontologies/prefixes included in the context
            #
            for cc in self.base_ontologies.items():
                if self._context.get(cc[0]) is None:
                    self._context[cc[0]] = cc[1]
            #
            # we always want the knora ontologies/prefixes included in the context
            #
            for cc in self.knora_ontologies.items():
                if self._context.get(cc[0]) is None:
                    self._context[cc[0]] = cc[1]
        else:
            self._context = {
                "rdf": OntoInfo("http://www.w3.org/1999/02/22-rdf-syntax-ns", True),
                "rdfs": OntoInfo("http://www.w3.org/2000/01/rdf-schema", True),
                "owl": OntoInfo("http://www.w3.org/2002/07/owl", True),
                "xsd": OntoInfo("http://www.w3.org/2001/XMLSchema", True),
                "knora-api": OntoInfo("http://api.knora.org/ontology/knora-api/v2", True),
                "salsah-gui": OntoInfo("http://api.knora.org/ontology/salsah-gui/v2", True)
            }
        self._rcontext = dict(map(lambda x: (x[1].iri, x[0]), self._context.items()))

    def __len__(self) -> int:
        return len(self._context)

    def __getitem__(self, key: str) -> OntoInfo:
        return self._context[key]

    def __setitem__(self, key: str, value: OntoInfo):
        self._context[key] = value
        self._rcontext[value.iri] = key

    def __delitem__(self, key: str) -> None:
        iri = self._context[key].iri
        del self._context[key].iri
        del self._rcontext[iri]

    def __contains__(self, key: str) -> bool:
        return key in self._context

    def __iter__(self) -> ContextIterator:
        return ContextIterator(self)

    def __str__(self) -> str:
        output = "Context:\n"
        for prefix, val in self._context.items():
            output += "  " + prefix + ": " + val.iri + "\n"
        return output

    #
    # now we have a lot of getters/setters
    #
    @property
    def context(self) -> ContextType:
        return self._context

    @context.setter
    def context(self, value: ContextType) -> None:
        """
        Setter function for the context out of a dict in the form { prefix1 : iri1, prefix2, iri2, …}

        :param value: Dictionary of context
        :return: None
        """
        if value is not None and isinstance(value, ContextType):
            self._context = value
        else:
            raise BaseError("Error in parameter to context setter")

    @property
    def rcontext(self) -> Dict[str, str]:
        return self._rcontext

    def add_context(self, prefix: str, iri: Optional[str] = None) -> None:
        """
        Add a new context to a context instance

        :param prefix: The prefix that should be used
        :param iri: The IRI that belongs to this context prefix
        :return: None
        """
        if iri is None:
            if prefix in self.knora_ontologies:
                return
            if prefix in self.base_ontologies:
                return
            if prefix in self.common_ontologies:
                self._context[prefix] = self.common_ontologies[prefix]
            else:
                raise BaseError("The prefix '{}' is not known!".format(prefix))
        else:
            if iri.endswith("#"):
                iri = iri[:-1]
                self._context[prefix] = OntoInfo(iri, True)
            else:
                self._context[prefix] = OntoInfo(iri, False)
        self._rcontext[iri] = prefix


    def iri_from_prefix(self, prefix: str) -> Optional[str]:
        """
        Returns the full IRI belonging to this prefix, without trailing "#"!

        :param prefix: Prefix of the context entry
        :return: The full IRI without trailing "#"
        """
        #if self.__is_iri(prefix):
        if IriTest.test(prefix):
            return prefix
        if self._context.get(prefix) is not None:
            return self._context.get(prefix).iri
        else:
            return None

    def prefix_from_iri(self, iri: str) -> Optional[str]:
        """
        Get the IRI from a full context that has or has not a a trailing "#". It first searches in the normal list
        of contexted. If the iri is not found there, it looks in the list common (external) ontologies. If the
        ontology is found there, this ontology is added to the list of known ontology and it's prefix is returned.
        If nothing is found, None is returns

        :param iri: The full IRI with or without trailing "#", or
        :return: the prefix of this context element, or None, if not found
        """
        #if not self.__is_iri(iri):
        if not IriTest.test(iri):
            raise BaseError("String does not conform to IRI patter: " + iri)
        if iri.endswith("#"):
            iri = iri[:-1]
        result = self._rcontext.get(iri)
        if result is None:
            entrylist = list(filter(lambda x: x[1].iri == iri, self.common_ontologies.items()))
            if len(entrylist) == 1:
                entry = entrylist[0]
                self._context[entry[0]] = entry[1]  # add to list of prefixes used
                self._rcontext[entry[1].iri] = entry[0]
                result = entry[0]
            else:
                tmp = iri.split('/')
                if tmp[-1] == "v2":
                    #
                    # we have a knora ontology name "http://server/ontology/shortcode/shortname/v2"
                    self._context[tmp[-2]] = OntoInfo(iri, True)  # add to list of prefixes used
                    self._rcontext[iri] = tmp[-2]
                else:
                    raise BaseError("Iri cannot be resolved to a well-known prefix!")
        return result

    def get_qualified_iri(self, val: Optional[str]) -> Optional[str]:
        """
        We will return the full qualified IRI, if it is not yet a full qualified IRI. If
        the IRI is already fully qualified, the we just return it.

        :param val: The input short form
        :return: the fully qualified IRI
        """
        if val is None:
            return None
        #if self.__is_iri(val):
        if IriTest.test(val):
            return val
        tmp = val.split(':')
        if len(tmp) < 2:
            raise BaseError("There is no separator to identify the prefix: " + val)
        iri_info = self._context.get(tmp[0])
        if iri_info is None:
            entrylist = list(filter(lambda x: x[1].iri == tmp[0], self.common_ontologies.items()))
            if len(entrylist) == 1:
                entry = entrylist[0]
                self._context[entry[0]] = entry[1]  # add to list of prefixes used
                self._rcontext[entry[1].iri] = entry[0]
                iri_info = entry[1]
            else:
                raise BaseError("Ontology not known! Cannot generate full qualified IRI")
        if iri_info.hashtag:
            return iri_info.iri + '#' + tmp[1]
        else:
            return iri_info.iri + tmp[1]

    def get_prefixed_iri(self, iri: Optional[str]) -> Optional[str]:
        """
        We reduce a fully qualified IRI to a short one in the form "prefix:name"

        :param iri: Fully qualified IRI
        :return: Return short from of IRI ("prefix:name")
        """

        if iri is None:
            return None
        #
        # check if the iri already has the form "prefix:name"
        #
        m = re.match("([\\w-]+):([\\w-]+)", iri)
        if m and m.span()[1] == len(iri):
            return iri
        #if not self.__is_iri(iri):
        if not IriTest.test(iri):
            raise BaseError("String does not conform to IRI patter: " + iri)

        splitpoint = iri.find('#')
        if splitpoint == -1:
            splitpoint = iri.rfind('/')
            ontopart = iri[:splitpoint + 1]
            element = iri[splitpoint + 1:]
        else:
            ontopart = iri[:splitpoint]
            element = iri[splitpoint + 1:]
        prefix = self._rcontext.get(ontopart)
        if prefix is None:
            entrylist = list(filter(lambda x: x[1].iri == ontopart, self.common_ontologies.items()))
            if len(entrylist) == 1:
                entry = entrylist[0]
                self._context[entry[0]] = entry[1]  # add to list of prefixes used
                self._rcontext[entry[1].iri] = entry[0]
                prefix = entry[0]
            else:
                raise BaseError("Ontology {} not known! Cannot generate full qualified IRI: prefix={}".format(iri, prefix))
        return prefix + ':' + element

    def reduce_iri(self, iristr: str, ontoname: Optional[str] = None) -> str:
        """
        Reduce an IRI to the form that is used within the definition json file. It expects
        the context object to have entries (prefixes)  for all IRI's
        - if it's an external IRI, it returns: "prefix:name"
        - if it's in the same ontology, it returns ":name"
        - if it's a system ontoloy ("knora-api" or "salsah-gui") it returns "name"
        :param iristr:
        :return:
        """
        rdf = self.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs = self.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = self.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        xsd = self.prefix_from_iri("http://www.w3.org/2001/XMLSchema#")
        knora_api = self.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui = self.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")

        #if self.__is_iri(iristr):
        if IriTest.test(iristr):
            iristr = self.get_prefixed_iri(iristr)
        tmp = iristr.split(':')
        if tmp[0] == knora_api or tmp[0] == salsah_gui:
            return tmp[1]
        elif ontoname is not None and tmp[0] == ontoname:
            return ':' + tmp[1]
        else:
            return iristr

    def toJsonObj(self) -> Dict[str, str]:
        """
        Return a python object that can be jsonfied...
        :return: Object to be jsonfied
        """
        return {prefix: oinfo.iri + '#' if oinfo.hashtag else oinfo.iri
                for prefix, oinfo in self._context.items()}

    def get_externals_used(self) -> Dict[str, str]:
        exclude = ["rdf", "rdfs", "owl", "xsd", "knora-api", "salsah-gui"]
        return {prefix: onto.iri for prefix, onto in self._context.items() if prefix not in exclude}

    def print(self) -> None:
        for a in self._context.items():
            print(a[0] + ': "' + a[1].iri + '"')


class LastModificationDate:
    """
    Class to hold and process the last modification date of a ontology
    """
    _last_modification_date: str;

    def __init__(self, val: Any):
        """
        The constructor works for different inputs:
        - a string holding the modification date
        - an instance of "LastModificationDate"
        - json-ld construct of the form { "@type": "xsd:dateTimeStamp", "@value": "date-str" }
        :param val: datetimestamp as string, instance of "LastModificationDate" or json-ld construct
        """
        if isinstance(val, str):
            self._last_modification_date = val;
        elif isinstance(val, LastModificationDate):
            self._last_modification_date = str(val)
        else:
            if val.get("@type") is not None and val.get("@type") == "xsd:dateTimeStamp":
                self._last_modification_date = val["@value"]
            else:
                raise BaseError("Invalid LastModificationDate")

    def __eq__(self, other: Union[str, 'LastModificationDate']) -> bool:
        if isinstance(other, str):
            other = LastModificationDate(other)
        return self._last_modification_date == other._last_modification_date

    def __lt__(self, other: 'LastModificationDate') -> bool:
        if isinstance(other, str):
            other = LastModificationDate(other)
        return self._last_modification_date < other._last_modification_date

    def __le__(self, other: 'LastModificationDate') -> bool:
        if isinstance(other, str):
            other = LastModificationDate(other)
        return self._last_modification_date <= other._last_modification_date

    def __gt__(self, other: 'LastModificationDate') -> bool:
        if isinstance(other, str):
            other = LastModificationDate(other)
        return self._last_modification_date > other._last_modification_date

    def __ge__(self, other: 'LastModificationDate') -> bool:
        if isinstance(other, str):
            other = LastModificationDate(other)
        return self._last_modification_date >= other._last_modification_date

    def __ne__(self, other: 'LastModificationDate') -> bool:
        if isinstance(other, str):
            other = LastModificationDate(other)
        return self._last_modification_date != other._last_modification_date

    def __str__(self: 'LastModificationDate') -> Union[None, str]:
        return self._last_modification_date

    def toJsonObj(self):
        return {
            "@type": "xsd:dateTimeStamp",
            "@value": self._last_modification_date
        }


class WithId:
    """
    Class helper to get json-ld "@id" thingies
    """
    _tmp: str = None

    def __init__(self, obj: Optional[Dict[str, str]]):
        if obj is None:
            return
        self._tmp = obj.get('@id')

    def str(self) -> Optional[str]:
        return self._tmp





import json
from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union, NewType
from urllib.parse import quote_plus

from .helpers import Actions, BaseError
from .langstring import Languages, LangStringParam, LangString
from .connection import Connection
from .model import Model
from .project import Project

from pprint import pprint

class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)

"""
This module implements the handling (CRUD) of Knora ListNodes and adds some function to read whole lists.

CREATE:
    * Instantiate a new object of the class ListNode with all required parameters
    * Call the ``create``-method on the instance

READ:
    * Instantiate a new object with ``id``(IRI of list(-node) given
    * Call the ``read``-method on the instance
    * Access the information that has been provided to the instance

UPDATE:
    * Only partially implemented. Only "label" and "comment" attributes may be changed.
    * You need an instance of an existing ListNode by reading an instance
    * Change the attributes by assigning the new values
    * Call the ``update```method on the instance

DELETE
    * NOT YET IMPLEMENTED BY Knora backend!
    * Instantiate a new objects with ``id``(IRI of project) given, or use any instance that has the id set
    * Call the ``delete``-method on the instance

In addition there is a static methods ``getAllProjects`` which returns a list of all projects
"""


@strict
class ListNode(Model):
    """
    This class represents a list node or a while list from Knora

    Attributes
    ----------

    con : Connection
        A Connection instance to a Knora server (for some operation a login has to be performedwith valid credentials)

    id : str
        IRI of the project [readonly, cannot be modified after creation of instance]

    project : str
        IRI of project. Only used for the creation of a new list (root node) [write].

    label : LangString
        A LangString instance with language depenedent labels. Setting this attributes overwites all entries
        with the new ones. In order to add/remove a specific entry, use "addLabel" or "rmLabel".
        At least one label is required [read/write].

    comment : LangString
        A LangString instance with language depenedent comments. Setting this attributes overwites all entries
        with the new ones.In order to add/remove a specific entry, use "addComment" or "rmComment".

    name : str
        A unique name for the ListNode (unique regarding the whole list) [read/write].

    parent : IRI | ListNode
        Is required and allowed only for the CREATE operation. Otherwise use the
        "children" attribute [write].

    isRootNode : bool
        Is True if the ListNode is a root node of a list Cannot be set [read].

    children : List[ListNode]
        Contains a list of children nodes. This attribute is only avaliable for nodes that have been read by the
        method "getAllNodes()"! [read]

    rootNodeIri : str
        IRI of the root node. This attribute is only avaliable for nodes that have been read by the
        method "getAllNodes()"! [read]

    Methods
    -------

    create : Knora ListNode information object
        Creates a new project and returns the information from the project as it is in Knora. Used to create new lists
        or append new ListNodes to an existing list. If appending, the attribute "parent" must not be None!

    read : Knora ListNode information object
        Read single list node

    update : Knora ListNode information object
        Updates the changed attributes and returns the updated information from the ListNode as it is in Knora

    delete : Knora result code
        Deletes a ListNode and returns the result code [NOT YET IMPLEMENTED!]

    getAllNodes : ListNode
        Get all nodes of a list. The IRI of the root node must be supplied.

    getAllLists [static]:
        returns all lists of a given project.

    print : None
        Prints the ListNode information to stdout (no recursion for children!)

    """

    _id: Union[str, None]
    _project: Union[str, None]
    _label: LangString
    _comments: LangString
    _name: Union[str, None]
    _parent: Union[str, None]
    _isRootNode: bool
    _children: Union[List['ListNode'], None]
    _rootNodeIri: Union[str, None]

    def __init__(self,
                 con: Connection,
                 id: Optional[str] = None,
                 project: Optional[Union[Project, str]] = None,
                 label: LangStringParam = None,
                 comments: LangStringParam = None,
                 name: Optional[str] = None,
                 parent: Optional[Union['ListNode', str]] = None,
                 isRootNode: Optional[bool] = None,
                 children: Optional[List['ListNode']] = None,
                 rootNodeIri: Optional[str] = None):
        """
        This is the constructor for the ListNode object. For

        CREATE:
            * The "con" and at least one "label" are required
        READ:
            * The "con" and "id" attributes are required
        UPDATE:
            * Only "label", "comments" and "name" may be changed
        DELETE:
            * Not yet implemented in the Knora-backend

        :param con: A valid Connection instance with a user logged in that has the appropriate permissions
        :param id: IRI of the project [readonly, cannot be modified after creation of instance]
        :param project: IRI of project. Only used for the creation of a new list (root node) [write].
        :param label: A LangString instance with language depenedent labels. Setting this attributes overwites all entries with the new ones. In order to add/remove a specific entry, use "addLabel" or "rmLabel". At least one label is required [read/write].
        :param comments: A LangString instance with language depenedent comments. Setting this attributes overwites all entries with the new ones.In order to add/remove a specific entry, use "addComment" or "rmComment".
        :param name: A unique name for the ListNode (unique regarding the whole list) [read/write].
        :param parent: Is required and allowed only for the CREATE operation. Otherwise use the "children" attribute [write].
        :param isRootNode: Is True if the ListNode is a root node of a list Cannot be set [read].
        :param children: Contains a list of children nodes. This attribute is only avaliable for nodes that have been read by the method "getAllNodes()"! [read]
        :param rootNodeIri: IRI of the root node. This attribute is only avaliable for nodes that have been read by the method "getAllNodes()"! [read]
        """

        super().__init__(con)

        self._project = project.id if isinstance(project, Project) else str(project) if project is not None else None
        self._id = str(id) if id is not None else None
        self._label = LangString(label)
        self._comments = LangString(comments)
        self._name = str(name) if name is not None else None
        if parent and isinstance(parent, ListNode):
            self._parent = parent.id
        else:
            self._parent = str(parent) if parent else None
        self._isRootNode = isRootNode
        if children is not None:
            if isinstance(children, List) and len(children) > 0 and isinstance(children[0], ListNode):
                self._children = children
            else:
                raise BaseError('Children must be list of ListNodes!')
        else:
            self._children = None
        if not isinstance(rootNodeIri, str) and rootNodeIri is not None:
            raise BaseError('rootNodeIri must be a str!')
        self._rootNodeIri = rootNodeIri

    #
    # Here follows a list of getters/setters
    #
    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        raise BaseError('ListNode id cannot be modified!')

    @property
    def project(self) -> Optional[str]:
        return self._project

    @project.setter
    def project(self, value: str) -> None:
        if self._project:
            raise BaseError('Project id cannot be modified!')
        else:
            self._project = value

    @property
    def label(self) -> Optional[LangString]:
        return self._label

    @label.setter
    def label(self, value: Optional[Union[LangString, str]]) -> None:
        self._label = LangString(value)
        self._changed.add('label')

    def addLabel(self, lang: Union[Languages, str], value: str) -> None:
        """
        Add/replace a node label with the given language (executed at next update)

        :param lang: The language the label, either a string "EN", "DE", "FR", "IT" or a Language instance
        :param value: The text of the label
        :return: None
        """

        self._label[lang] = value
        self._changed.add('label')

    def rmLabel(self, lang: Union[Languages, str]) -> None:
        """
        Remove a label from a list node (executed at next update)

        :param lang: The language the label to be removed is in, either a string "EN", "DE", "FR", "IT" or a Language instance
        :return: None
        """
        del self._label[lang]
        self._changed.add('label')

    @property
    def comments(self) -> Optional[LangString]:
        return self._comments

    @comments.setter
    def comments(self,  value: Optional[Union[LangString, str]]) -> None:
        self._comments = LangString(value)
        self._changed.add('comments')

    def addComment(self, lang: Union[Languages, str], value: str) -> None:
        """
        Add/replace a node comments with the given language (executed at next update)

        :param lang: The language the comments, either a string "EN", "DE", "FR", "IT" or a Language instance
        :param value: The text of the comments
        :return: None
        """

        self._comments[lang] = value
        self._changed.add('comments')

    def rmComment(self, lang: Union[Languages, str]) -> None:
        """
        Remove a comments from a list node (executed at next update)

        :param lang: The language the comment to be removed is in, either a string "EN", "DE", "FR", "IT" or a Language instance
        :return: None
        """
        del self._comments[lang]
        self._changed.add('comments')

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = str(value)
        self._changed.add('name')

    @property
    def parent(self) -> Optional[str]:
        return self._parent if self._parent else None

    @parent.setter
    def parent(self, value: Union[str, 'ListNode']):
        if isinstance(value, ListNode):
            self._parent = value.id
        else:
            self._parent = value

    @property
    def isRootNode(self) -> Optional[bool]:
        return self._isRootNode

    @isRootNode.setter
    def isRootNode(self, value: bool) -> None:
        raise BaseError('Property isRootNode cannot be set!')

    @property
    def children(self) -> Optional[List['ListNode']]:
        return self._children

    @staticmethod
    def __getChildren(con: Connection,
                      parent_iri: str,
                      project_iri: str,
                      children: List[Any]) -> Optional[List['ListNode']]:
        """
        Internal method! Should not be used directly!

        Static method gets a recursive List of children nodes

        :param con: Valid Connection instance
        :param children: json object of children
        :return: List of ListNode instances
        """
        if children:
            child_nodes: List[Any] = []
            for child in children:

                if 'parentNodeIri' not in child:
                    child['parentNodeIri'] = parent_iri
                if 'projectIri' not in child:
                    child['projectIri'] = project_iri
                child_node = ListNode.fromJsonObj(con, child)
                child_nodes.append(child_node)
            return child_nodes
        else:
            return None

    @property
    def rootNodeIri(self) -> Optional[str]:
        return self._rootNodeIri

    @rootNodeIri.setter
    def rootNodeIri(self, value: str):
        raise BaseError('rootNodeIri cannot be set!')

    def has_changed(self) -> bool:
        if self._changed:
            return True
        else:
            return False

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any) -> Any:
        """
        Internal method! Should not be used directly!

        This method is used to create a ListNode instance from the JSON data returned by Knora

        :param con: Connection instance
        :param json_obj: JSON data returned by Knora as python3 object
        :return: ListNode instance
        """

        id = json_obj.get('id')
        if id is None:
            raise BaseError('ListNode id is missing')
        project = json_obj.get('projectIri')
        label = LangString.fromJsonObj(json_obj.get('labels'))
        comments = LangString.fromJsonObj(json_obj.get('comments'))
        if json_obj.get('name') is not None:
            name = json_obj['name']
        else:
            name = id.rsplit('/', 1)[-1]
        parent = json_obj.get('parentNodeIri')
        isRootNode = json_obj.get('isRootNode')

        child_info = json_obj.get('children')
        children = None
        if child_info is not None:
            children = ListNode.__getChildren(con=con,
                                              parent_iri=id,
                                              project_iri=project,
                                              children=child_info)
        rootNodeIri = json_obj.get('hasRootNode')

        return cls(con=con,
                   id=id,
                   project=project,
                   label=label,
                   comments=comments,
                   name=name,
                   parent=parent,
                   isRootNode=isRootNode,
                   children=children,
                   rootNodeIri=rootNodeIri)

    def toJsonObj(self, action: Actions, listIri: str = None) -> Any:
        """
        Internal method! Should not be used directly!

        Creates a JSON-object from the ListNode instance that can be used to call Knora

        :param action: Action the object is used for (Action.CREATE or Action.UPDATE)
        :return: JSON-object
        """

        tmp = {}
        if action == Actions.Create:
            if self._project is None:
                raise BaseError("There must be a project id given!")
            tmp['projectIri'] = self._project
            if self._label.isEmpty():
                raise BaseError("There must be a valid ListNode label!")
            tmp['labels'] = self._label.toJsonObj()
            if not self._comments.isEmpty():
                tmp['comments'] = self._comments.toJsonObj()
            else:
                tmp['comments'] = []
            if self._name is not None:
                tmp['name'] = self._name
            if self._parent is not None:
                tmp['parentNodeIri'] = self._parent
        elif action == Actions.Update:
            if self.id is None:
                raise BaseError("There must be a node id given!")
            tmp['listIri'] = listIri
            if self._project is None:
                raise BaseError("There must be a project id given!")
            tmp['projectIri'] = self._project
            if not self._label.isEmpty() and 'label' in self._changed:
                tmp['labels'] = self._label.toJsonObj()
            if not self._comments.isEmpty() and 'comments' in self._changed:
                tmp['comments'] = self._comments.toJsonObj()
            if self._name is not None and 'name' in self._changed:
                tmp['name'] = self._name
        return tmp

    def create(self) -> 'ListNode':
        """
        Create a new List in Knora

        :return: JSON-object from Knora
        """

        jsonobj = self.toJsonObj(Actions.Create)
        jsondata = json.dumps(jsonobj, cls=SetEncoder)
        if self._parent is not None:
            result = self._con.post('/admin/lists/' + quote_plus(self._parent), jsondata)
            return ListNode.fromJsonObj(self._con, result['nodeinfo'])
        else:
            result = self._con.post('/admin/lists', jsondata)
            return ListNode.fromJsonObj(self._con, result['list']['listinfo'])

    def read(self) -> Any:
        """
        Read a project from Knora

        :return: JSON-object from Knora
        """

        result = self._con.get('/admin/lists/nodes/' + quote_plus(self._id))
        if result.get('nodeinfo'):
            return self.fromJsonObj(self._con, result['nodeinfo'])
        elif result.get('listinfo'):
            return self.fromJsonObj(self._con, result['listinfo'])
        else:
            return None

    def update(self) -> Union[Any, None]:
        """
        Udate the ListNode info in Knora with the modified data in this ListNode instance

        :return: JSON-object from Knora refecting the update
        """

        jsonobj = self.toJsonObj(Actions.Update, self.id)
        if jsonobj:
            jsondata = json.dumps(jsonobj, cls=SetEncoder)
            result = self._con.put('/admin/lists/' + quote_plus(self.id), jsondata)
            pprint(result)
            return ListNode.fromJsonObj(self._con, result['listinfo'])
        else:
            return None

    def delete(self) -> None:
        """
        Delete the given ListNode

        :return: Knora response
        """
        raise BaseError("NOT YET IMPLEMENTED BY KNORA BACKEND!")
        result = self._con.delete('/admin/lists/' + quote_plus(self._id))
        return result
        #return Project.fromJsonObj(self.con, result['project'])

    def getAllNodes(self) -> 'ListNode':
        """
        Get all nodes of the list. Must be called from a ListNode instance that has at least set the
        list iri!

        :return: Root node of list with recursive ListNodes ("children"-attributes)
        """

        result = self._con.get('/admin/lists/' + quote_plus(self._id))
        if 'list' not in result:
            raise BaseError("Request got no list!")
        if 'listinfo' not in result['list']:
            raise BaseError("Request got no proper list information!")
        root = ListNode.fromJsonObj(self._con, result['list']['listinfo'])
        if 'children' in result['list']:
            root._children = ListNode.__getChildren(con=self._con,
                                                    parent_iri=root.id,
                                                    project_iri=root.project,
                                                    children=result['list']['children'])
        return root

    @staticmethod
    def getAllLists(con: Connection, project_iri: Optional[str] = None) -> List['ListNode']:
        """
        Get all lists. If a project IRI is given, it returns the lists of the specified project

        :param con: Connection instance
        :param project_iri: Iri/id of project
        :return: list of ListNodes
        """
        if project_iri is None:
            result = con.get('/admin/lists')
        else:
            result = con.get('/admin/lists?projectIri=' + quote_plus(project_iri))
        if 'lists' not in result:
            raise BaseError("Request got no lists!")
        return list(map(lambda a: ListNode.fromJsonObj(con, a), result['lists']))

    def _createDefinitionFileObj(self, children: List["ListNode"]):
        """
        Create an object that corresponds to the syntax of the input to "create_onto".
        Node: This method must be used only internally (for recursion)!!

        :param children: List of children nodes
        :return: A python object that can be jsonfied to correspond to the syntax of the input to "create_onto".
        """
        listnodeobjs = []
        for listnode in children:
            listnodeobj = {
                "name": listnode.name,
                "labels": listnode.label.createDefinitionFileObj(),
            }
            if not listnode.comments.isEmpty():
                listnodeobj["comments"] = listnode.comments.createDefinitionFileObj()
            if listnode.children:
                listnodeobj["nodes"] = self._createDefinitionFileObj(listnode.children)
            listnodeobjs.append(listnodeobj)
        return listnodeobjs

    def createDefinitionFileObj(self):
        """
        Create an object that corresponds to the syntax of the input to "create_onto".
        :return: A python object that can be jsonfied to correspond to the syntax of the input to "create_onto".
        """
        listnode = {
            "name": self._name,
            "labels": self._label.createDefinitionFileObj(),
        }
        if not self._comments.isEmpty():
            listnode["comments"] = self._comments.createDefinitionFileObj()
        if self._children:
            listnode["nodes"] = self._createDefinitionFileObj(self._children)
        return listnode

    def print(self):
        """
        print info to stdout

        :return: None
        """

        print('Node Info:')
        print('  Id:        {}'.format(self._id))
        print('  Project:   {}'.format(self._project))
        print('  Name:      {}'.format(self._name))
        print('  Label:     ')
        if self._label:
            for lbl in self._label.items():
                print('             {}: {}'.format(lbl[0], lbl[1]))
        else:
            print('             None')
        print('  Comments:   ')
        if self._comments.isEmpty():
            for lbl in self._comments.items():
                print('             {}: {}'.format(lbl[0], lbl[1]))
        else:
            print('             None')
        print('  Parent', self._parent)
        print('  IsRootNode: {}'.format(self._isRootNode))



import os
import json
from jsonschema import validate

from .onto_commons import validate_list_from_excel, json_list_from_excel

from pprint import pprint

def validate_list(input_file: str) -> None:
    current_dir = os.path.dirname(os.path.realpath(__file__))

    # let's read the schema for the data model definition
    with open(os.path.join(current_dir, 'knora-schema-lists.json')) as s:
        schema = json.load(s)
    # read the data model definition
    with open(input_file) as f:
        datamodel = json.load(f)

    # validate the data model definition in order to be sure that it is correct
    validate(datamodel, schema)
    print("Data model is syntactically correct and passed validation!")


def validate_ontology(input_file: str) -> None:
    with open(input_file) as f:
        jsonstr = f.read()
    validate_ontology_from_string(jsonstr)


def validate_ontology_from_string(jsonstr: str) -> None:
    current_dir = os.path.dirname(os.path.realpath(__file__))

    with open(os.path.join(current_dir, 'knora-schema.json')) as s:
        schema = json.load(s)
    # read the data model definition
    #with open(input_file) as f:
    #    datamodel = json.load(f)
    #with open(input_file) as f:
    #    jsonstr = f.read()
    datamodel = json.loads(jsonstr)

    #
    # now let's see if there are any lists defined as reference to excel files
    #
    lists = datamodel["project"].get('lists')
    if lists is not None:
        newlists: [] = []
        for rootnode in lists:
            if rootnode.get("nodes") is not None and isinstance(rootnode["nodes"], dict) and rootnode["nodes"].get("file") is not None:
                newroot = {
                    "name": rootnode.get("name"),
                    "labels": rootnode.get("labels")
                }
                if rootnode.get("comments") is not None:
                    newroot["comments"] = rootnode["comments"]
                startrow = 1 if rootnode["nodes"].get("startrow") is None else rootnode["nodes"]["startrow"]
                startcol = 1 if rootnode["nodes"].get("startcol") is None else rootnode["nodes"]["startcol"]
                json_list_from_excel(rootnode=newroot,
                                     filepath=rootnode["nodes"]["file"],
                                     sheetname=rootnode["nodes"]["worksheet"],
                                     startrow=startrow,
                                     startcol=startcol)
                newlists.append(newroot)
            else:
                newlists.append(rootnode)
        datamodel["project"]["lists"] = newlists

    json.dump(datamodel, open("gaga.json", "w"), indent=4)

    # validate the data model definition in order to be sure that it is correct
    validate(datamodel, schema)

    print("Data model is syntactically correct and passed validation!")





import json
import re
from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union
from enum import Enum
from urllib.parse import quote_plus

from pprint import pprint

from .helpers import Actions, BaseError, Context, Cardinality, LastModificationDate
from .connection import Connection
from .model import Model
from .langstring import Languages, LangString

class SetEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)

"""
This model implements the handling of resource classes. It contains two classes that work closely together:
    * "HasProperty" deals with the association of Property-instances with the Resource-instances. This association
      is done using the "cardinality"-clause
    * "ResourceClass" is the main class representing a knora resource class.
"""
@strict
class HasProperty(Model):
    class Ptype(Enum):
        system = 1
        knora = 2
        other = 3

    _context: Context
    _ontology_id: str
    _property_id: str
    _resclass_id: str
    _cardinality: Cardinality
    _gui_order: int
    _is_inherited: bool
    _ptype: Ptype

    def __init__(self,
                 con: Connection,
                 context: Context,
                 ontology_id: Optional[str] = None,
                 property_id: Optional[str] = None,
                 resclass_id: Optional[str] = None,
                 cardinality: Optional[Cardinality] = None,
                 gui_order: Optional[int] = None,
                 is_inherited: Optional[bool] = None,
                 ptype: Optional[Ptype] = None):
        super().__init__(con)
        if not isinstance(context, Context):
            raise BaseError('"context"-parameter must be an instance of Context')
        self._context = context
        if ontology_id is not None:
            self._ontology_id = context.iri_from_prefix(ontology_id)
        else:
            self._ontology_id = None
        self._property_id = property_id
        self._resclass_id = resclass_id
        self._cardinality = cardinality
        self._gui_order = gui_order
        self._is_inherited = is_inherited
        self._ptype = ptype
        self._changed = set()

    #
    # Here follows a list of getters/setters
    #
    @property
    def ontology_id(self) -> Optional[str]:
        return self._ontology_id

    @ontology_id.setter
    def ontology_id(self, value: Optional[str]) -> None:
        self._ontology_id = value

    @property
    def property_id(self) -> Optional[str]:
        return self._property_id

    @property_id.setter
    def property_id(self, value: str) -> None:
        raise BaseError('property_id "{}" cannot be modified!'.format(self._property_id))

    @property
    def resclass_id(self) -> Optional[str]:
        return self._resclass_id

    @resclass_id.setter
    def resclass_id(self, value: Optional[str]) -> None:
        self._resclass_id = value

    @property
    def cardinality(self) -> Optional[Cardinality]:
        return self._cardinality

    @cardinality.setter
    def cardinality(self, value: Optional[Cardinality]) -> None:
        self._cardinality = value
        self._changed.add('cardinality')

    @property
    def gui_order(self) -> Optional[int]:
        return self._gui_order

    @gui_order.setter
    def gui_order(self, value: int) -> None:
        self._gui_order = value
        self._changed.add('gui_order')

    @property
    def ptype(self) -> Ptype:
        return self._ptype

    @classmethod
    def fromJsonObj(cls, con: Connection, context: Context, jsonld_obj: Any) -> Tuple[str, 'HasProperty']:
        if not isinstance(con, Connection):
            raise BaseError('"con"-parameter must be an instance of Connection')
        if not isinstance(context, Context):
            raise BaseError('"context"-parameter must be an instance of Context')

        rdf = context.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        xsd = context.prefix_from_iri("http://www.w3.org/2001/XMLSchema#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui = context.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")

        if jsonld_obj.get('@type') is None or jsonld_obj.get('@type') != owl + ":Restriction":
            raise BaseError('Expected restriction type')

        #
        # let's get the inherited field...
        #
        tmp = jsonld_obj.get(knora_api + ':isInherited')
        is_inherited = tmp if tmp is not None else False

        #
        # let's get the cardinality
        #
        cardinality: Cardinality
        if jsonld_obj.get(owl + ':cardinality') is not None:
            cardinality = Cardinality.C_1
        elif jsonld_obj.get(owl + ':maxCardinality') is not None:
            cardinality = Cardinality.C_0_1
        elif jsonld_obj.get(owl + ':minCardinality') is not None:
            if jsonld_obj.get(owl + ':minCardinality') == 0:
                cardinality = Cardinality.C_0_n
            elif jsonld_obj.get(owl + ':minCardinality') == 1:
                cardinality = Cardinality.C_1_n
            else:
                raise BaseError('Problem with cardinality')
        else:
            raise BaseError('Problem with cardinality')

        #
        # Now let's get the property IRI
        #
        property_id: str
        ptype: HasProperty.Ptype
        ontology_id: Optional[str] = None
        if jsonld_obj.get(owl + ':onProperty') is None:
            raise BaseError('No property IRI given')
        p = jsonld_obj[owl + ':onProperty'].get('@id')
        if p is None:
            raise BaseError('No property IRI given')
        pp = p.split(':')
        if pp[0] == rdf or pp[0] == rdfs or pp[0] == owl:
            ptype = HasProperty.Ptype.system
        elif pp[0] == knora_api:
            ptype = HasProperty.Ptype.knora
        else:
            ptype = HasProperty.Ptype.other
            ontology_id = context.iri_from_prefix(pp[0])
        property_id = p

        gui_order: int = None
        if jsonld_obj.get(salsah_gui + ':guiOrder') is not None:
            gui_order = jsonld_obj[salsah_gui + ':guiOrder']
        return property_id, cls(con=con,
                                context=context,
                                ontology_id=ontology_id,
                                property_id=property_id,
                                cardinality=cardinality,
                                gui_order=gui_order,
                                is_inherited=is_inherited,
                                ptype=ptype)

    def toJsonObj(self, lastModificationDate: LastModificationDate, action: Actions) -> Any:
        if self._cardinality is None:
            raise BaseError("There must be a cardinality given!")
        tmp = {}
        switcher = {
            Cardinality.C_1: ("owl:cardinality", 1),
            Cardinality.C_0_1: ("owl:maxCardinality", 1),
            Cardinality.C_0_n: ("owl:minCardinality", 0),
            Cardinality.C_1_n: ("owl:minCardinality", 1)
        }
        occurrence = switcher.get(self._cardinality)
        if action == Actions.Create:
            tmp = {
                "@id": self._ontology_id,
                "@type": "owl:Ontology",
                "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
                "@graph": [{
                    "@id": self._resclass_id,
                    "@type": "owl:Class",
                    "rdfs:subClassOf": {
                        "@type": "owl:Restriction",
                        occurrence[0]: occurrence[1],
                        "owl:onProperty": {
                            "@id": self._property_id
                        }
                    }
                }],
                "@context": self._context.toJsonObj()
            }
            if self._gui_order is not None:
                tmp["@graph"][0]["rdfs:subClassOf"]["salsah-gui:guiOrder"] = self._gui_order
        elif action == Actions.Update:
            tmp = {
                "@id": self._ontology_id,
                "@type": "owl:Ontology",
                "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
                "@graph": [{
                    "@id": self._resclass_id,
                    "@type": "owl:Class",
                    "rdfs:subClassOf": {
                        "@type": "owl:Restriction",
                        occurrence[0]: occurrence[1],
                        "owl:onProperty": {
                            "@id": self._property_id
                        }
                    }
                }],
                "@context": self._context.toJsonObj()
            }
            if self._gui_order is not None and 'gui_order' in self._changed:
                tmp["@graph"][0]["rdfs:subClassOf"]["salsah-gui:guiOrder"] = self._gui_order
        return tmp

    def create(self, last_modification_date: LastModificationDate) -> Tuple[LastModificationDate, 'ResourceClass']:
        if self._ontology_id is None:
            raise BaseError("Ontology id required")
        if self._property_id is None:
            raise BaseError("Property id required")
        if self._cardinality is None:
            raise BaseError("Cardinality id required")

        jsonobj = self.toJsonObj(last_modification_date, Actions.Create)
        jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=2)
        result = self._con.post('/v2/ontologies/cardinalities', jsondata)
        last_modification_date = LastModificationDate(result['knora-api:lastModificationDate'])
        return last_modification_date, ResourceClass.fromJsonObj(self._con, self._context, result['@graph'])

    def update(self, last_modification_date: LastModificationDate) -> Tuple[LastModificationDate, 'ResourceClass']:
        if self._ontology_id is None:
            raise BaseError("Ontology id required")
        if self._property_id is None:
            raise BaseError("Property id required")
        if self._cardinality is None:
            raise BaseError("Cardinality id required")
        jsonobj = self.toJsonObj(last_modification_date, Actions.Update)
        jsondata = json.dumps(jsonobj, indent=3, cls=SetEncoder)
        result = self._con.put('/v2/ontologies/cardinalities', jsondata)
        last_modification_date = LastModificationDate(result['knora-api:lastModificationDate'])
        # TODO: self._changed = str()
        return last_modification_date, ResourceClass.fromJsonObj(self._con, self._context, result['@graph'])

    def delete(self, last_modification_date: LastModificationDate) -> LastModificationDate:
        raise BaseError("Cannot remove a single property from a class!")
        # ToDo: Check with Ben if we could add this feature...

    def createDefinitionFileObj(self, context: Context, shortname: str):
        cardinality = {}
        if self._ptype == HasProperty.Ptype.other:
            cardinality["propname"] = context.reduce_iri(self._property_id, shortname)
            cardinality["cardinality"] = self._cardinality.value
            if self._gui_order is not None:
                cardinality["gui_order"] = self._gui_order
        return cardinality

    def print(self, offset: int = 0):
        blank = ' '
        if self._ptype == HasProperty.Ptype.system:
            print(f'{blank:>{offset}}Has Property (system)')
        elif self._ptype == HasProperty.Ptype.knora:
            print(f'{blank:>{offset}}Has Property (knora)')
        else:
            print(f'{blank:>{offset}}Has Property (project)')
        print(f'{blank:>{offset + 2}}Property: {self._property_id}')
        print(f'{blank:>{offset + 2}}Cardinality: {self._cardinality.value}')
        if self._ptype == HasProperty.Ptype.other:
            print(f'{blank:>{offset + 2}}Ontology_id: {self._ontology_id}')
        print(f'{blank:>{offset + 2}}Resclass: {self._resclass_id}')


@strict
class ResourceClass(Model):
    """
    This class represents a knora resource class

    Attributes
    ----------

    con : Connection
        A Connection instance to a Knora server

    id : str
        IRI of the ResourceClass [readonly, cannot be modified after creation of instance]

    name: str
        The name of the resource class, e.g. "Book", "Person", "Portait". Usually these names start
        with a capital letter

    ontology_id: str
        The IRI/Id of the ontology this resource class belongs to

    superlcasses: str, List[str]
        This is a list of superclasses for this resource class. Usually a project specific class must at least
        be a subclass of "Resource", but can be subclassed of any other valid resource class. In addition, external
        ontologies may be referenced:
        e.g.:
        ```
        "super": "Resource"

        "super": ":MySpecialTYpe"

        "super": [":MySpecialType", "dcterms:BibliographicResource"]
        ```

    label: language dependent string, that is a dict like {"en": "Biblography", "de": "Literaturverzeichnis"}
        A label (human readable name) for the resource

    comment: language dependent string, that is a dict like {"en": "a comment", "de": "Ein Kommentar"}
        A commentary to further explain what this resource class is used for

    permission: str
        The default permissions to be used if an instance of this resource class is being created

    has_properties: Dict[str, HasProperty]
        Holds a dict with the property names as keys and a HasProperty instance. HasProperty holds
        the information, how this resource class uses this property (basically the cardinality)

    changed: bool
        is set to True, if one of the fields has been chaned by the user (internal use only!)


    Methods
    -------

    addLabel: Add a new label to the resource
        addLabel(self, lang: Union[Languages, str], value: str) -> None

    getProperty: Get information about a property defined for this resource class
        getPropertery(prop_id: str) -> HasProperty
        returns a HasProperty-instance

    addProperty: Add a new property to the resource class
        addProperty(property_id: str, cardinality: Cardinality, last_modification_date: LastModificationDate)
        -> Optional[LastModificationDate]

    updateProperty: Updates the cardinality parameters of the given property with the resource class
        updateProperty(self, property_id: str, cardinality: Cardinality, last_modification_date: LastModificationDate)
        -> Optional[LastModificationDate]
        Please note that the cardinality usually can only be changed to be *less* restrictive!

    create: Create a new resource class on the connected server

    update: Update the information of a resource class on the connected server

    delete: Mark a resource class as deleted (on the connected server)

    createDefinitionFileObj: Create an object suitable for jsonification that conforms the the knora-py ontology tools

    print: Print the content of this class to the console

    """
    _context: Context
    _id: str
    _name: str
    _ontology_id: str
    _superclasses: List[str]
    _label: LangString
    _comment: LangString
    _permissions: str
    _has_properties: Dict[str, HasProperty]

    def __init__(self,
                 con: Connection,
                 context: Context,
                 id: Optional[str] = None,
                 name: Optional[str] = None,
                 ontology_id: Optional[str] = None,
                 superclasses: Optional[List[Union['ResourceClass', str]]] = None,
                 label: Optional[Union[LangString, str]] = None,
                 comment: Optional[Union[LangString, str]] = None,
                 permissions: Optional[str] = None,
                 has_properties: Optional[Dict[str, HasProperty]] = None):
        """
        Create an instance of  ResourceClass

        :param con:
        :param context:
        :param id:
        :param name:
        :param ontology_id:
        :param superclasses:
        :param label:
        :param comment:
        :param permissions:
        :param has_properties:
        """
        super().__init__(con)
        if not isinstance(con, Connection):
            raise BaseError('"con"-parameter must be an instance of Connection')
        if not isinstance(context, Context):
            raise BaseError('"context"-parameter must be an instance of Context')
        self._context = context
        self._id = id
        self._name = name
        if ontology_id is not None:
            self._ontology_id = context.iri_from_prefix(ontology_id)
        if isinstance(superclasses, ResourceClass):
            self._superclasses = list(map(lambda a: a.id, superclasses))
        else:
            self._superclasses = superclasses
        #
        # process label
        #
        if label is not None:
            if isinstance(label, str):
                self._label = LangString(label)
            elif isinstance(label, LangString):
                self._label = label
            else:
                raise BaseError('Invalid LangString for label!')
        else:
            self._label = None
        #
        # process comment
        #
        if comment is not None:
            if isinstance(comment, str):
                self._comment = LangString(comment)
            elif isinstance(comment, LangString):
                self._comment = comment
            else:
                raise BaseError('Invalid LangString for comment!')
        else:
            self._comment = None
        self._permissions = permissions
        self._has_properties = has_properties
        self._changed = set()

    #
    # Here follows a list of getters/setters
    #
    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        raise BaseError('"name" cannot be modified!')

    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        raise BaseError('"id" cannot be modified!')

    @property
    def ontology_id(self) -> Optional[str]:
        return self._ontology_id

    @ontology_id.setter
    def ontology_id(self, value: str) -> None:
        raise BaseError('"ontology_id" cannot be modified!')

    @property
    def superclasses(self) -> Optional[List[str]]:
        return self._superclasses

    @superclasses.setter
    def superclasses(self, value: List[str]) -> None:
        raise BaseError('"superclasses" cannot be modified!')

    @property
    def label(self) -> Optional[LangString]:
        return self._label

    @label.setter
    def label(self, value: Optional[Union[LangString, str]]) -> None:
        if value is None:
            self._label.empty()  # clear all labels
        else:
            if isinstance(value, LangString):
                self._label = value
            elif isinstance(value, str):
                self._label = LangString(value)
            else:
                raise BaseError('Not a valid LangString')
        self._changed.add('label')


    def addLabel(self, lang: Union[Languages, str], value: str) -> None:
        self._label[lang] = value
        self._changed.add('label')

    def rmLabel(self, lang: Union[Languages, str]) -> None:
        del self._label[lang]
        self._changed.add('label')

    @property
    def comment(self) -> Optional[LangString]:
        return self._comment

    @comment.setter
    def comment(self, value: Optional[LangString]) -> None:
        if value is None:
            self._comment.empty()  # clear all comments!
        else:
            if isinstance(value, LangString):
                self._comment = value
            elif isinstance(value, str):
                self._comment = LangString(value)
            else:
                raise BaseError('Not a valid LangString')
        self._changed.add('comment')

    def addComment(self, lang: Union[Languages, str], value: str) -> None:
        self._comment[lang] = value
        self._changed.add('comment')

    def rmComment(self, lang: Union[Languages, str]) -> None:
        del self._comment[lang]
        self._changed.add('comment')

    @property
    def permissions(self) -> Optional[str]:
        return self._permissions

    @permissions.setter
    def permissions(self, value: str) -> None:
        raise BaseError('"permissions" cannot be modified!')

    @property
    def has_properties(self) -> Dict[str, HasProperty]:
        return self._has_properties

    @has_properties.setter
    def has_properties(self, value: str) -> None:
        raise BaseError('"has_property" cannot be modified!')

    def getProperty(self, property_id) -> Optional[HasProperty]:
        if self._has_properties is None:
            return None
        else:
            return self._has_properties.get(self._context.get_prefixed_iri(property_id))

    def addProperty(self,
                    last_modification_date: LastModificationDate,
                    property_id: str,
                    cardinality: Cardinality,
                    gui_order: Optional[int] = None,
                    ) -> Optional[LastModificationDate]:
        if self._has_properties.get(property_id) is None:
            latest_modification_date, resclass = HasProperty(con=self._con,
                                                             context=self._context,
                                                             ontology_id=self._ontology_id,
                                                             property_id=property_id,
                                                             resclass_id=self.id,
                                                             cardinality=cardinality,
                                                             gui_order=gui_order).create(last_modification_date)
            hp = resclass.getProperty(property_id)
            hp._ontology_id = self._context.iri_from_prefix(self._ontology_id)
            hp._resclass_id = self._id
            self._has_properties[hp.property_id] = hp
            return latest_modification_date
        else:
            raise BaseError("Property already has cardinality in this class! " + property_id)

    def updateProperty(self,
                       last_modification_date: LastModificationDate,
                       property_id: str,
                       cardinality: Optional[Cardinality],
                       gui_order: Optional[int] = None,
                       ) -> Optional[LastModificationDate]:
        property_id = self._context.get_prefixed_iri(property_id)
        if self._has_properties.get(property_id) is not None:
            has_properties = self._has_properties[property_id]
            #onto_id = has_properties.ontology_id  # save for later user
            #rescl_id = has_properties.resclass_id  # save for later user
            has_properties.ontology_id = self._ontology_id
            has_properties.resclass_id = self._id
            if cardinality:
                has_properties.cardinality = cardinality
            if gui_order:
                has_properties.gui_order = gui_order
            latest_modification_date, resclass = has_properties.update(last_modification_date)
            hp = resclass.getProperty(property_id)
            hp.ontology_id = self._ontology_id  # self.__context.iri_from_prefix(onto_id)  # restore value
            hp.resclass_id = self._id  # rescl_id  # restore value
            self._has_properties[hp.property_id] = hp
            return latest_modification_date
        else:
            return last_modification_date

    @classmethod
    def fromJsonObj(cls, con: Connection, context: Context, json_obj: Any) -> Any:
        if isinstance(json_obj, list):
            json_obj = json_obj[0]  # TODO: Is it possible to have more than one element in the list??
        if not isinstance(con, Connection):
            raise BaseError('"con"-parameter must be an instance of Connection')
        if not isinstance(context, Context):
            raise BaseError('"context"-parameter must be an instance of Context')
        rdf = context.prefix_from_iri("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
        rdfs = context.prefix_from_iri("http://www.w3.org/2000/01/rdf-schema#")
        owl = context.prefix_from_iri("http://www.w3.org/2002/07/owl#")
        xsd = context.prefix_from_iri("http://www.w3.org/2001/XMLSchema#")
        knora_api = context.prefix_from_iri("http://api.knora.org/ontology/knora-api/v2#")
        salsah_gui = context.prefix_from_iri("http://api.knora.org/ontology/salsah-gui/v2#")

        if not (json_obj.get(knora_api + ':isResourceClass') or json_obj.get(knora_api + ':isStandoffClass')):
            raise BaseError("This is not a resource!")

        if json_obj.get('@id') is None:
            raise BaseError('Resource class has no "@id"!')
        tmp_id = json_obj.get('@id').split(':')
        id = context.iri_from_prefix(tmp_id[0]) + '#' + tmp_id[1]
        ontology_id = tmp_id[0]
        name = tmp_id[1]
        superclasses_obj = json_obj.get(rdfs + ':subClassOf')
        if superclasses_obj is not None:
            supercls: List[Any] = list(filter(lambda a: a.get('@id') is not None, superclasses_obj))
            superclasses: List[str] = list(map(lambda a: a['@id'], supercls))
            has_props: List[Any] = list(filter(lambda a: a.get('@type') == (owl + ':Restriction'), superclasses_obj))
            has_properties: Dict[HasProperty] = dict(map(lambda a: HasProperty.fromJsonObj(con, context, a), has_props))
            #
            # now remove the ...Value stuff from resource pointers: A resource pointer is returned as 2 properties:
            # one a direct link, the other the pointer to a link value
            #
            tmp = dict(has_properties)
            for key in tmp.keys():
                key_with_value = key
                if key.endswith("Value"):
                    key = key.removesuffix("Value")
                    if key in has_properties:
                        del has_properties[key_with_value]
        else:
            superclasses = None
            has_properties = None

        label = LangString.fromJsonLdObj(json_obj.get(rdfs + ':label'))
        comment = LangString.fromJsonLdObj(json_obj.get(rdfs + ':comment'))
        return cls(con=con,
                   context=context,
                   name=name,
                   id=id,
                   ontology_id=ontology_id,
                   superclasses=superclasses,
                   label=label,
                   comment=comment,
                   has_properties=has_properties)

    def toJsonObj(self, lastModificationDate: LastModificationDate, action: Actions, what: Optional[str] = None) -> Any:

        def resolve_resref(resref: str):
            tmp = resref.split(':')
            if len(tmp) > 1:
                if tmp[0]:
                    self._context.add_context(tmp[0])
                    return {"@id": resref}  # fully qualified name in the form "prefix:name"
                else:
                    return {"@id": self._context.prefix_from_iri(self._ontology_id) + ':' + tmp[1]}  # ":name" in current ontology
            else:
                return {"@id": "knora-api:" + resref}  # no ":", must be from knora-api!

        tmp = {}
        exp = re.compile('^http.*')  # It is already a fully IRI
        if exp.match(self._ontology_id):
            resid = self._context.prefix_from_iri(self._ontology_id) + ":" + self._name
            ontid = self._ontology_id
        else:
            resid = self._ontology_id + ":" + self._name
            ontid = self._context.iri_from_prefix(self._ontology_id)
        if action == Actions.Create:
            if self._name is None:
                raise BaseError("There must be a valid resource class name!")
            if self._ontology_id is None:
                raise BaseError("There must be a valid ontology_id given!")
            if self._superclasses is None:
                superclasses = [{"@id": "knora-api:Resource"}]
            else:
                superclasses = list(map(resolve_resref, self._superclasses))
            if self._comment is None or self._comment.isEmpty():
                self._comment = LangString("no comment available")
            if self._label is None or self._label.isEmpty():
                self._label = LangString("no label available")
            tmp = {
                "@id": ontid,  # self._ontology_id,
                "@type": "owl:Ontology",
                "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
                "@graph": [{
                    "@id": resid,
                    "@type": "owl:Class",
                    "rdfs:label": self._label.toJsonLdObj(),
                    "rdfs:comment": self._comment.toJsonLdObj(),
                    "rdfs:subClassOf": superclasses
                }],
                "@context": self._context.toJsonObj(),
            }
        elif action == Actions.Update:
            tmp = {
                "@id": ontid,  # self._ontology_id,
                "@type": "owl:Ontology",
                "knora-api:lastModificationDate": lastModificationDate.toJsonObj(),
                "@graph": [{
                    "@id": resid,
                    "@type": "owl:Class",
                }],
                "@context": self._context.toJsonObj(),
            }
            if what == 'label':
                if not self._label.isEmpty() and 'label' in self._changed:
                    tmp['@graph'][0]['rdfs:label'] = self._label.toJsonLdObj()
            if what == 'comment':
                if not self._comment.isEmpty() and 'comment' in self._changed:
                    tmp['@graph'][0]['rdfs:comment'] = self._comment.toJsonLdObj()
        return tmp

    def create(self, last_modification_date: LastModificationDate) -> Tuple[LastModificationDate, 'ResourceClass']:
        jsonobj = self.toJsonObj(last_modification_date, Actions.Create)
        jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=4)
        result = self._con.post('/v2/ontologies/classes', jsondata)
        last_modification_date = LastModificationDate(result['knora-api:lastModificationDate'])
        return last_modification_date, ResourceClass.fromJsonObj(self._con, self._context, result['@graph'])

    def update(self, last_modification_date: LastModificationDate) -> Tuple[LastModificationDate, 'ResourceClass']:
        #
        # Note: Knora is able to change only one thing per call, either label or comment!
        #
        result = None
        something_changed = False
        if 'label' in self._changed:
            jsonobj = self.toJsonObj(last_modification_date, Actions.Update, 'label')
            jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=4)
            result = self._con.put('v2/ontologies/classes', jsondata)
            last_modification_date = LastModificationDate(result['knora-api:lastModificationDate'])
            something_changed = True
        if 'comment' in self._changed:
            jsonobj = self.toJsonObj(last_modification_date, Actions.Update, 'comment')
            jsondata = json.dumps(jsonobj, cls=SetEncoder, indent=4)
            result = self._con.put('v2/ontologies/classes', jsondata)
            last_modification_date = LastModificationDate(result['knora-api:lastModificationDate'])
            something_changed = True
        if something_changed:
            return last_modification_date, ResourceClass.fromJsonObj(self._con, self._context, result['@graph'])
        else:
            return last_modification_date, self

    def delete(self, last_modification_date: LastModificationDate) -> LastModificationDate:
        result = self._con.delete('v2/ontologies/classes/' + quote_plus(self._id) + '?lastModificationDate=' + str(last_modification_date))
        return LastModificationDate(result['knora-api:lastModificationDate'])

    def createDefinitionFileObj(self, context: Context, shortname: str, skiplist: List[str]):
        resource = {
            "name": self._name,
            "labels": self._label.createDefinitionFileObj(),
        }
        if self._comment:
            resource["comments"] = self._comment.createDefinitionFileObj()
        if self._superclasses:
            if len(self._superclasses) > 1:
                superclasses = []
                for sc in self._superclasses:
                    superclasses.append(context.reduce_iri(sc, shortname))
            else:
                superclasses = context.reduce_iri(self._superclasses[0], shortname)
            resource["super"] = superclasses
        if self._has_properties:
            cardinalities = []
            for pid, hp in self._has_properties.items():
                if hp.property_id in skiplist:
                    continue
                if hp.ptype == HasProperty.Ptype.other:
                    cardinalities.append(hp.createDefinitionFileObj(context, shortname))
            resource["cardinalities"] = cardinalities

        return resource

    def print(self, offset: int = 0):
        blank = ' '
        print(f'{blank:>{offset}}Resource Class Info')
        print(f'{blank:>{offset+2}}Name:            {self._name}')
        print(f'{blank:>{offset+2}}Ontology prefix: {self._ontology_id}')
        print(f'{blank:>{offset+2}}Superclasses:')
        if self._superclasses is not None:
            for sc in self._superclasses:
                print(f'{blank:>{offset + 4}}{sc}')
        if self._label is not None:
            print(f'{blank:>{offset + 2}}Labels:')
            self._label.print(offset + 4)
        if self._comment is not None:
            print(f'{blank:>{offset + 2}}Comments:')
            self._comment.print(offset + 4)
        if self._has_properties is not None:
            print(f'{blank:>{offset + 2}}Has properties:')
            if self._has_properties is not None:
                for pid, hp in self._has_properties.items():
                    hp.print(offset + 4)








import json
from pystrict import strict
from typing import List, Set, Dict, Tuple, Optional, Any, Union
from urllib.parse import quote_plus

from .helpers import Actions, BaseError
from .connection import Connection
from .model import Model
from .project import Project

"""
This module implements the handling (CRUD) of Knora groups.

CREATE:
    * Instantiate a new object of the class Group with all required parameters
    * Call the ``create``-method on the instance

READ:
    * Instantiate a new object with ``id``(IRI of group) given
    * Call the ``read``-method on the instance
    * Access the information that has been provided to the instance

UPDATE:
    * You need an instance of an existing Project by reading an instance
    * Change the attributes by assigning the new values
    * Call the ``update```method on the instance

DELETE
    * Instantiate a new objects with ``id``(IRI of group) given, or use any instance that has the id set
    * Call the ``delete``-method on the instance

"""
@strict
class Group(Model):
    """
    This class represents a Knora group

    Attributes
    ----------

    con : Connection
        A connection instance to a Knora server

    id : str
        IRI of the group [get only, cannot be modified after creation of instance]

    name : str
        Name of the group

    description : str
        A description of the group

    project : str | project
        either the IRI of a project [get only, cannot be modified after creation of instance]
        or an valid Project instance

    selfjoin : boolean
        A flag indicating if selfjoin is allowed in this group

    status : boolean
        A flag indicating if the group is active (True) or inactive/makred deleted (False)

    """

    PROJECT_MEMBER_GROUP: str = "http://www.knora.org/ontology/knora-admin#ProjectMember"
    PROJECT_ADMIN_GROUP: str = "http://www.knora.org/ontology/knora-admin#ProjectAdmin"
    PROJECT_SYSTEMADMIN_GROUP: str = "http://www.knora.org/ontology/knora-admin#SystemAdmin"

    _id: str
    _name: str
    _description: str
    _project: str
    _selfjoin: bool
    _status: bool

    def __init__(self,
                 con: Connection,
                 id: Optional[str] = None,
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 project: Optional[Union[str, Project]] = None,
                 selfjoin: Optional[bool] = None,
                 status: Optional[bool] = None):
        super().__init__(con)
        self._id = str(id) if id is not None else None
        self._name = str(name) if name is not None else None
        self._description = description
        if project is not None and isinstance(project, Project):
            self._project = project.id
        else:
            self._project = str(project) if project is not None else None
        self._selfjoin = bool(selfjoin) if selfjoin is not None else None
        self._status = bool(status) if status is not None else None

    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        raise BaseError('Group id cannot be modified!')

    @property
    def name(self) -> Optional[str]:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value
        self._changed.add('name')

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value: Optional[str]):
        self._description = value
        self._changed.add('description')

    @property
    def project(self):
        return self._project

    @project.setter
    def project(self, value: str):
        raise BaseError('project id cannot be modified!')

    @property
    def selfjoin(self) -> bool:
        return self._selfjoin

    @selfjoin.setter
    def selfjoin(self, value: bool) -> None:
        self._selfjoin = value
        self._changed.add('selfjoin')

    @property
    def status(self) -> bool:
        return self._status

    @status.setter
    def status(self, value: bool) -> None:
        self._status = value
        self._changed.add('status')

    def has_changed(self) -> bool:
        if self._changed:
            return True
        else:
            return False

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any):
        id = json_obj.get('id')
        if id is None:
            raise BaseError('Group "id" is missing in JSON from knora')
        name = json_obj.get('name')
        if name is None:
            raise BaseError('Group "name" is missing in JSON from knora')
        description = json_obj.get('description')
        tmp = json_obj.get('project')
        if tmp is None:
            raise BaseError('Group "project" is missing in JSON from knora')
        project = tmp.get('id')
        if project is None:
            raise BaseError('Group "project" has no "id" in JSON from knora')
        selfjoin = json_obj.get('selfjoin')
        if selfjoin is None:
            raise BaseError("selfjoin is missing in JSON from knora")
        status = json_obj.get('status')
        if status is None:
            raise BaseError("Status is missing in JSON from knora")
        return cls(con=con,
                   name=name,
                   id=id,
                   description=description,
                   project=project,
                   selfjoin=selfjoin,
                   status=status)

    def toJsonObj(self, action: Actions):
        tmp = {}
        if action == Actions.Create:
            if self._name is None:
                raise BaseError("There must be a valid name!")
            tmp['name'] = self._name
            if self._description is not None:
                tmp['description'] = self._description
            if self._project is None:
                raise BaseError("There must be a valid project!")
            tmp['project'] = self._project
            if self._selfjoin is None:
                raise BaseError("There must be a valid value for selfjoin!")
            tmp['selfjoin'] = self._selfjoin
            if self._status is None:
                raise BaseError("There must be a valid value for status!")
            tmp['status'] = self._status
        else:
            if self._name is not None and 'name' in self._changed:
                tmp['name'] = self._name
            if self._description is not None and 'description' in self._changed:
                tmp['description'] = self._description
            if self._selfjoin is not None and 'selfjoin' in self._changed:
                tmp['selfjoin'] = self._selfjoin
        return tmp

    def create(self):
        jsonobj = self.toJsonObj(Actions.Create)
        jsondata = json.dumps(jsonobj)
        result = self._con.post('/admin/groups', jsondata)
        return Group.fromJsonObj(self._con, result['group'])

    def read(self):
        result = self._con.get('/admin/groups/' + quote_plus(self._id))
        return Group.fromJsonObj(self._con, result['group'])

    def update(self):
        jsonobj = self.toJsonObj(Actions.Update)
        if jsonobj:
            jsondata = json.dumps(jsonobj)
            result = self._con.put('/admin/groups/' + quote_plus(self._id), jsondata)
            updated_group = Group.fromJsonObj(self._con, result['group'])
        if self._status is not None and 'status' in self._changed:
            jsondata = json.dumps({'status': self._status})
            result = self._con.put('/admin/groups/' + quote_plus(self._id) + '/status', jsondata)
            updated_group = Group.fromJsonObj(self._con, result['group'])
        return updated_group

    def delete(self):
        result = self._con.delete('/admin/groups/' + quote_plus(self._id))
        return Group.fromJsonObj(self._con, result['group'])

    @staticmethod
    def getAllGroups(con: Connection) -> List['Group']:
        result = con.get('/admin/groups')
        if 'groups' not in result:
            raise BaseError("Request got no groups!")
        return list(map(lambda a: Group.fromJsonObj(con, a), result['groups']))


    def print(self):
        print('Group Info:')
        print('  Id:          {}'.format(self._id))
        print('  Name:        {}'.format(self._name))
        print('  Description: {}'.format(self._description))
        print('  Project:     {}'.format(self._project))
        print('  Selfjoin:    {}'.format(self._selfjoin))
        print('  Status:      {}'.format(self._status))



if __name__ == '__main__':
    con = Connection('http://0.0.0.0:3333')
    con.login('root@example.com', 'test')

    groups = Group.getAllGroups(con)
    for group in groups:
        group.print()

    new_group = Group(con=con,
                           name="KNORA-PY TEST",
                           description="Test project for knora-py",
                           project="http://rdfh.ch/projects/00FF",
                           status=True,
                           selfjoin=False).create()
    new_group.print()

    new_group.name="KNORA-PY TEST - modified"
    new_group = new_group.update();
    new_group.print()

    new_group.description="gaga gaga gaga gaga gaga gaga gaga"
    new_group = new_group.update();
    new_group.print()

    new_group.selfjoin = True
    new_group = new_group.update();
    new_group.print()

    new_group.status = False
    new_group = new_group.update();
    new_group.print()

    new_group.name = '-- KNORA-PY TEST --'
    new_group.description = 'Final Test'
    new_group.status = True
    new_group = new_group.update()
    new_group.print()

    new_group.delete()

    print('=========================')
    groups = Group.getAllGroups(con)
    for group in groups:
        group.print()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='dsp-tools',
    version='1.0.4',
    description='A Python library and tools for the DaSCH Service Platform',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/dasch-swiss/dsp-tools',
    author='Lukas Rosenthaler',
    author_email='lukas.rosenthaler@dasch.swiss',
    license='GPLv3',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9.0',
    install_requires=[
        'argparse',
        'rdflib',
        'lxml',
        'validators',
        'requests',
        'jsonschema',
        'click',
        'rfc3987',
        'pystrict',
        'openpyxl'
    ],
    entry_points={
          'console_scripts': [
              'dsp-tools=knora.dsp_tools:main'
          ],
    },
    include_package_data=True,
    zip_safe=False,
)



import unittest
import json

from dsplib.utils.onto_get import get_ontology
from dsplib.utils.onto_get import get_ontology
from dsplib.utils.onto_validate import validate_ontology
from dsplib.utils.onto_create_ontology import create_ontology
from knora.dsplib.utils.xml_upload import xml_upload


class TestTools(unittest.TestCase):

    def test_get(self):

        with open('testdata/anything.json') as f:
            jsonstr = f.read()
        refobj = json.loads(jsonstr)

        get_ontology(projident="anything",
                     outfile="_anything.json",
                     server="http://0.0.0.0:3333",
                     user="root@example.com",
                     password="test",
                     verbose=True)

        with open('_anything.json') as f:
            jsonstr = f.read()
        jsonobj = json.loads(jsonstr)

        self.assertEqual(refobj["project"]["shortcode"], jsonobj["project"]["shortcode"])

    def test_validate_onto(self):
        validate_ontology('testdata/test-onto.json')

    def test_create_onto(self):
        create_ontology(input_file='testdata/test-onto.json',
                        lists_file='lists-out.json',
                        server="http://0.0.0.0:3333",
                        user="root@example.com",
                        password="test",
                        verbose=True,
                        dump=True)

    def test_xmlupload(self):
        xml_upload(input_file="testdata/test-data.xml",
                   server="http://0.0.0.0:3333",
                   user="root@example.com",
                   password="test",
                   imgdir="testdata/bitmaps",
                   sipi="http://0.0.0.0:1024",
                   verbose=True,
                   validate=True)



if __name__ == '__main__':
    unittest.main()



from typing import List, Set, Dict, Tuple, Optional, Any, Union

import json
import re

from ..models.connection import Connection
from ..models.project import Project
from ..models.listnode import ListNode
from ..models.ontology import Ontology


def get_ontology(projident: str, outfile: str, server: str, user: str, password: str, verbose: bool) -> bool:
    con = Connection(server)
    # con.login(user, password)
    if re.match("^[0-9aAbBcCdDeEfF]{4}$", projident):
        project = Project(con=con, shortcode=projident)
    elif re.match("^[\\w-]+$", projident):
        project = Project(con=con, shortname=projident)
    elif re.match("^(http)s?://([\\w\\.\\-~]+:?\\d{,4})(/[\\w\\-~]+)+$", projident):
        project = Project(con=con, shortname=projident)
    else:
        print("Invalid project identification!")
        return False

    project = project.read()

    projectobj = project.createDefinitionFileObj()

    #
    # now collect the lists
    #
    listroots = ListNode.getAllLists(con=con, project_iri=project.id)
    listobj = []
    for listroot in listroots:
        complete_list = listroot.getAllNodes()
        listobj.append(complete_list.createDefinitionFileObj())
    projectobj["lists"] = listobj

    projectobj["ontologies"] = []
    prefixes: Dict[str, str] = {}
    ontologies = Ontology.getProjectOntologies(con, project.id)
    ontology_ids = [x.id for x in ontologies]
    for ontology in ontology_ids:
        oparts = ontology.split("/")
        name = oparts[len(oparts) - 2]
        shortcode = oparts[len(oparts) - 3]
        ontology = Ontology.getOntologyFromServer(con=con, shortcode=shortcode, name=name)
        projectobj["ontologies"].append(ontology.createDefinitionFileObj())
        prefixes.update(ontology.context.get_externals_used())

    umbrella = {
        "prefixes": prefixes,
        "project": projectobj
    }

    with open(outfile, 'w', encoding='utf8') as outfile:
        json.dump(umbrella, outfile, indent=3, ensure_ascii=False)




from typing import List, Set, Dict, Tuple, Optional, Any, Union
from enum import Enum, unique

from ..models.helpers import BaseError


@unique
class Languages(Enum):
    EN = 'en'
    DE = 'de'
    FR = 'fr'
    IT = 'it'

LangStringParam = Optional[Union[Dict[Union[Languages, str], str], str]]


class LangStringIterator:
    """Iterator class for LangString class."""

    _langstring: 'LangString'
    _index: int

    def __init__(self, langstring: 'LangString'):
        self._langstring = langstring;
        self._langlist = list(map(lambda a: a[0], self._langstring.items()))
        self._index = 0

    def __next__(self):
        if len(self._langlist) == 0 and self._index == 0:
            return None, self._langstring[None]
        elif self._index < len(self._langlist):
            tmp = self._langlist[self._index]
            self._index += 1
            return tmp, self._langstring[tmp]
        else:
            raise StopIteration


class LangString:
    """
    This class holds a typical language-string pair as it is often used in JSON-LD as
    ```
    "some:thing" = [{
      '@language': 'xx',
      '@value': 'a string in language xx'
    },
    {…},…]
    ```

    or a simple string without language:
    ```
    "some:thing": "a string without language specificer"
    ```
    """
    _langstrs: Dict[Languages, str]
    _simplestring: str

    def __init__(self, initvalue: LangStringParam = None):

        def mymapper(p: Tuple[Union[Languages, str], str]) -> Tuple[Languages, str]:
            lmap = dict(map(lambda a: (a.value, a), Languages))
            if isinstance(p[0], str) and p[0] in lmap:
                lang = lmap[p[0].lower()]
            elif isinstance(p[0], Languages):
                lang = p[0]
            else:
                raise BaseError("No a valid language definition!")
            return lang, p[1]

        if initvalue is None:
            self._simplestring = None
            self._langstrs = {}
        elif isinstance(initvalue, str):
            self._simplestring = initvalue
            self._langstrs = {}
        elif isinstance(initvalue, dict):
            self._simplestring = None
            self._langstrs = dict(map(mymapper, initvalue.items()))
        elif isinstance(initvalue, LangString):
            self._simplestring = initvalue._simplestring
            self._langstrs = initvalue._langstrs
        else:
            raise BaseError("No a valid language definition!")

    def __getitem__(self, key: Optional[Union[Languages, str]] = None) -> str:
        #
        # First deal with simple strings (no language given). We return, if existing, the simple string
        # or just the first language dependent string
        #
        if key is None:
            if self._simplestring is not None:
                return self._simplestring
            else:
                try:
                    return list(self._langstrs)[0]
                except:
                    return None
        else:
            pass
            # self._simplestring = None  # Let's delete the string without language if there is one...
        if isinstance(key, Enum):
            if self._langstrs.get(key) is None:
                for l in self._langstrs:
                    if self._langstrs.get(l) is not None:
                        return self._langstrs[l]
                if self._simplestring is not None:
                    return self._simplestring
                return None
            else:
                return self._langstrs[key]
        else:
            lmap = dict(map(lambda a: (a.value, a), Languages))
            if lmap.get(key.lower()) is None:
                raise BaseError('Invalid language string "' + key + '"!')
            if self._langstrs.get(lmap[key.lower()]) is None:
                for l in self._langstrs:
                    if self._langstrs.get(l) is not None:
                        return self._langstrs[l]
                if self._simplestring is not None:
                    return self._simplestring
                return None
            else:
                return self._langstrs[lmap[key.lower()]]

    def __setitem__(self, key: Optional[Union[Languages, str]], value: str):
        if key is None:
            self._simplestring = value
            self._langstrs = {}  # Delete language dependent string! Needs Caution!!!
            return
        if isinstance(key, Languages):
            self._langstrs[key] = value
        else:
            lmap = dict(map(lambda a: (a.value, a), Languages))
            if lmap.get(key.lower()) is None:
                raise BaseError('Invalid language string "' + key  + '"!')
            self._langstrs[lmap[key.lower()]] = value

    def __delitem__(self, key: Union[Languages, str]):
        if isinstance(key, Languages):
            del self._langstrs[key]
        else:
            lmap = dict(map(lambda a: (a.value, a), Languages))
            if lmap.get(key.lower()) is None:
                raise BaseError('Invalid language string "' + key  + '"!')
            del self._langstrs[lmap[key.lower()]]

    def __str__(self):
        tmpstr = '{'
        for p in self._langstrs:
            tmpstr += '=' + p.value + ':' + self._langstrs[p]
        tmpstr += '}'
        return tmpstr

    def __iter__(self):
        return LangStringIterator(self)

    def __eq__(self, other: 'LangString'):
        equal = self._simplestring == other._simplestring
        if equal:
            for lang in Languages:
                equal = self._langstrs.get(lang) == other._langstrs.get(lang)
                if not equal:
                    break
        return equal


    def get_by_lang(self, key: Optional[Union[Languages, str]] = None) -> Optional[str]:
        if key is None:
            return self._simplestring
        else:
            if isinstance(key, Enum):
                return self._langstrs.get(key)
            else:
                lmap = dict(map(lambda a: (a.value, a), Languages))
                if lmap.get(key.lower()) is None:
                    raise BaseError('Invalid language string "' + key + '"!')
                return self._langstrs.get(lmap[key.lower()])

    def items(self):
        return self._langstrs.items()

    def isEmpty(self):
        return not (bool(self._langstrs) or bool(self._simplestring))

    def empty(self):
        self._simplestring = None
        self._langstrs = {}


    def toJsonObj(self):
        if self.isEmpty():
            return None
        if self._simplestring is not None:
            return [self._simplestring]
        else:
            return list(map(lambda a: {'language': a[0].value, 'value': a[1]}, self._langstrs.items()))

    def toJsonLdObj(self):
        if self.isEmpty():
            return None
        if self._simplestring is not None:
            return [self._simplestring]
        else:
            return list(map(lambda a: {'@language': a[0].value, '@value': a[1]}, self._langstrs.items()))

    @classmethod
    def fromJsonLdObj(cls, obj: Optional[Union[List[Dict[str, str]], str]]) -> 'LangString':
        if obj is None:
            return None
        if isinstance(obj, str):
            return cls(obj)
        if isinstance(obj, list):
            objs = obj
        else:
            objs = [obj]
        lstrs: Dict[Languages, str] = {}
        for o in objs:
            lang = o.get('@language')
            if lang == 'en':
                lstrs[Languages.EN] = o.get('@value')
            elif lang == 'de':
                lstrs[Languages.DE] = o.get('@value')
            elif lang == 'fr':
                lstrs[Languages.FR] = o.get('@value')
            elif lang == 'it':
                lstrs[Languages.IT] = o.get('@value')
            else:
                if o.get('@value') is not None:
                    return cls(o.get('@value'))
        return cls(lstrs)

    @classmethod
    def fromJsonObj(cls, obj: Optional[Any]) -> 'LangString':
        if obj is None:
            return None
        if isinstance(obj, str):
            return cls(obj)
        if isinstance(obj, list):
            objs = obj
        else:
            objs = [obj]
        lstrs: Dict[Languages, str] = {}
        for o in objs:
            lang = o.get('language')
            if lang == 'en':
                lstrs[Languages.EN] = o.get('value')
            elif lang == 'de':
                lstrs[Languages.DE] = o.get('value')
            elif lang == 'fr':
                lstrs[Languages.FR] = o.get('value')
            elif lang == 'it':
                lstrs[Languages.IT] = o.get('value')
            else:
                if o.get('value') is not None:
                    return cls(o.get('value'))
        return cls(lstrs)

    def print(self, offset: Optional[int] = None):
        blank = ' '
        #print(f'{blank:>{offset}}LangString:')
        if self._simplestring is not None:
            print(f'{blank:>{offset + 2}}{self._simplestring}')
        else:
            for p in self._langstrs.items():
                print(f'{blank:>{offset + 2}}{p[0]} : {p[1]}')

    @property
    def langstrs(self):
        return self._langstrs

    def createDefinitionFileObj(self) -> Union[str, Dict[str, str]]:
        if self._simplestring is not None:
            return self._simplestring
        langstring = {}
        for p in self.items():
            langstring[p[0].value] = p[1]
        return langstring



import unittest

from dsplib.models.connection import Connection
from dsplib.models.resource import ResourceInstanceFactory
from dsplib.models.value import BooleanValue, ColorValue, DateValue, DecimalValue, IntValue, IntervalValue, TextValue, \
    UriValue, KnoraStandoffXml, make_value
from dsplib.models.permission import PermissionValue, Permissions, PermissionsIterator
from dsplib.models.sipi import Sipi

from pprint import pprint

class TestResource(unittest.TestCase):

    def test_Resource_create(self):
        #
        # Connect to Knora
        #
        con = Connection('http://0.0.0.0:3333')
        con.login('root@example.com', 'test')

        #
        # Make class factory for project 'anything. The factory creates classes that implement the CRUD methods
        # for the given resource classes, that is to create, read, update and delete instances (=resources) of the given classesd
        #
        factory = ResourceInstanceFactory(con, 'anything')
        resclassnames = factory.get_resclass_names()

        #
        # Get an python class of a BlueThing resource class
        #
        BlueThing = factory.get_resclass('anything:BlueThing')

        an_old_thing = BlueThing(con=con, iri="http://rdfh.ch/0001/H6gBWUuJSuuO-CilHV8kQw").read()

        a_blue_thing = BlueThing(con=con,
                                 label='BlueThing',
                                 values={
                                     'anything:hasBoolean': True,
                                     'anything:hasColor': ['#ff0033', '#0077FF'],
                                     'anything:hasDate': make_value(value='1833', comment='not verified!'),
                                     'anything:hasDecimal': 3.14159,
                                     'anything:hasInteger': "42",
                                     'anything:hasTimeStamp': "2004-04-12T13:20:00Z",
                                     'anything:hasInterval': [make_value(value="0.0:3.0", comment="first interval"),
                                                              make_value(value="3.5:3.7", comment="second interval")],
                                     'anything:hasRichtext': KnoraStandoffXml("This is <em>bold</em> text."),
                                     'anything:hasText': "Dies ist ein einfacher Text",
                                     'anything:hasUri': 'http://gaga.com:65500/gugus'
                                 }).create()


        new_blue_thing = a_blue_thing.read()
        self.assertEqual(a_blue_thing.label, "BlueThing")
        self.assertEqual(a_blue_thing.value("anything:hasColor"), ['#ff0033', '#0077FF'])
        self.assertEqual(a_blue_thing.value("anything:hasInteger"), 42)
        self.assertEqual(a_blue_thing.value("anything:hasBoolean"), True)
        self.assertEqual(a_blue_thing.value("anything:hasDecimal"), 3.14159)
        self.assertEqual(a_blue_thing.value("anything:hasText"), "Dies ist ein einfacher Text")

        ThingPicture = factory.get_resclass('anything:ThingPicture')
        sipi = Sipi('http://0.0.0.0:1024', con.get_token())
        img = sipi.upload_bitstream('testdata/bitstreams/TEMP11.TIF')
        fileref = img['uploadedFiles'][0]['internalFilename']
        resperm = Permissions({PermissionValue.M: ["knora-admin:UnknownUser", "knora-admin:KnownUser"],
                       PermissionValue.CR: ["knora-admin:Creator", "knora-admin:ProjectAdmin"]})
        a_thing_picture = ThingPicture(
            con=con,
            label='ThingPicture',
            bitstream=fileref,
            permissions=resperm,
            values={
                'anything:hasPictureTitle': make_value(value="A Thing Picture named Lena", permissions=resperm)
            }
        ).create()


if __name__ == '__main__':
    unittest.main()



import os
from typing import List, Set, Dict, Tuple, Optional
import json
from jsonschema import validate

from ..models.helpers import Actions, BaseError, Context, Cardinality
from ..models.langstring import Languages, LangStringParam, LangString
from ..models.connection import Connection
from ..models.project import Project
from ..models.listnode import ListNode
from ..models.group import Group
from ..models.user import User
from ..models.ontology import Ontology
from ..models.propertyclass import PropertyClass
from ..models.resourceclass import ResourceClass

from .onto_commons import list_creator, validate_list_from_excel, json_list_from_excel

from pprint import pprint

def create_ontology(input_file: str,
                    lists_file: str,
                    server: str,
                    user: str,
                    password: str,
                    verbose: bool,
                    dump: bool) -> bool:
    current_dir = os.path.dirname(os.path.realpath(__file__))

    # let's read the schema for the data model definition
    with open(os.path.join(current_dir, 'knora-schema.json')) as s:
        schema = json.load(s)


    # read the data model definition
    with open(input_file) as f:
        datamodel = json.load(f)

    #
    # now let's see if there are any lists defined as reference to excel files
    #
    lists = datamodel["project"].get('lists')
    if lists is not None:
        newlists: [] = []
        for rootnode in lists:
            if rootnode.get("nodes") is not None and isinstance(rootnode["nodes"], dict) and rootnode["nodes"].get("file") is not None:
                newroot = {
                    "name": rootnode.get("name"),
                    "labels": rootnode.get("labels"),
               }
                if rootnode.get("comments") is not None:
                    newroot["comments"] = rootnode["comments"]

                startrow = 1 if rootnode["nodes"].get("startrow") is None else rootnode["nodes"]["startrow"]
                startcol = 1 if rootnode["nodes"].get("startcol") is None else rootnode["nodes"]["startcol"]
                json_list_from_excel(rootnode=newroot,
                                     filepath=rootnode["nodes"]["file"],
                                     sheetname=rootnode["nodes"]["worksheet"],
                                     startrow=startrow,
                                     startcol=startcol)
                newlists.append(newroot)
            else:
                newlists.append(rootnode)
        datamodel["project"]["lists"] = newlists

    # validate the data model definition in order to be sure that it is correct
    validate(datamodel, schema)

    if verbose:
        print("Data model is syntactically correct and passed validation!")

    #
    # Connect to the DaSCH Service Platform API
    #
    con = Connection(server)
    con.login(user, password)
    if dump:
        con.start_logging()

    # --------------------------------------------------------------------------
    # let's read the prefixes of external ontologies that may be used
    #
    context = Context(datamodel["prefixes"])


    # --------------------------------------------------------------------------
    # Let's create the project...
    #
    project = None
    try:
        # we try to read the project to see if it's existing....
        project = Project(
            con=con,
            shortcode=datamodel["project"]["shortcode"],
        ).read()
        #
        # we got it, update the project data if necessary...
        #
        if project.shortname != datamodel["project"]["shortname"]:
            project.shortname = datamodel["project"]["shortname"]
        if project.longname != datamodel["project"]["longname"]:
            project.longname == datamodel["project"]["longname"]
        project.description = datamodel["project"].get("descriptions")
        project.keywords = set(datamodel["project"].get("keywords"))
        nproject = project.update()
        if nproject is not None:
            project = nproject
        if verbose:
            print("Modified project:")
            project.print()
    except:
        #
        # The project doesn't exist yet – let's create it
        #
        try:
            project = Project(
                con=con,
                shortcode=datamodel["project"]["shortcode"],
                shortname=datamodel["project"]["shortname"],
                longname=datamodel["project"]["longname"],
                description=LangString(datamodel["project"].get("descriptions")),
                keywords=set(datamodel["project"].get("keywords")),
                selfjoin=False,
                status=True
            ).create()
        except BaseError as err:
            print("Creating project failed: " + err.message)
            return False
        if verbose:
            print("Created project:")
            project.print()
    assert project is not None

    # --------------------------------------------------------------------------
    # now let's create the lists
    #
    if verbose:
        print("Creating lists...")
    lists = datamodel["project"].get('lists')
    listrootnodes = {}
    if lists is not None:
        for rootnode in lists:
            if verbose is not None:
                print("  Creating list:" + rootnode['name'])
            root_list_node = ListNode(
                con=con,
                project=project,
                label=rootnode['labels'],
                comments=rootnode.get('comments'),
                name=rootnode['name']
            ).create()
            listnodes = list_creator(con, project, root_list_node, rootnode['nodes'])
            listrootnodes[rootnode['name']] = {
                "id": root_list_node.id,
                "nodes": listnodes
            }

    with open('lists.json', 'w', encoding="utf-8") as fp:
        json.dump(listrootnodes, fp, indent=3, sort_keys=True)
        print("The definitions of the node-id's can be found in \"{}\"!".format('lists.json'))

    # --------------------------------------------------------------------------
    # now let's add the groups (if there are groups defined...)
    #
    if verbose:
        print("Adding groups...")

    new_groups = {}
    groups = datamodel["project"].get('groups')
    if groups is not None:
        for group in groups:
            try:
                new_group = Group(con=con,
                                  name=group["name"],
                                  description=group["description"],
                                  project=project,
                                  status=group["status"] if group.get("status") is not None else True,
                                  selfjoin=group["selfjoin"] if group.get("selfjoin") is not None else False).create()
            except BaseError as err:
                print("Creating group failed: " + err.message)
                return False
            new_groups[new_group.name] = new_group
            if verbose:
                print("Groups:")
                new_group.print()
            #project.set_default_permissions(new_group.id)
    # --------------------------------------------------------------------------
    # now let's add the users (if there are users defined...)
    #
    if verbose:
        print("Adding users...")
    all_groups: List[Group] = []
    all_projects: List[Project] = []
    users = datamodel["project"].get('users')
    if users is not None:
        for user in users:
            sysadmin = False
            group_ids: Set[str] = set()
            for groupname in user["groups"]:
                #
                # First we determine the groups the user is in because we can do this in one call
                # groupname has the form [proj_shortname]:groupname|"SystemAdmin"
                # (projectname omitted = current project)
                #
                tmp = groupname.split(':')
                if len(tmp) > 1:
                    group = None
                    if tmp[0] and tmp[0] != '':
                        # we have 'proj_shortname:groupname
                        if not all_groups:
                            all_groups = Group.getAllGroups(con)
                        tmp_group = list(filter(lambda g: g.project.shortname == tmp[0] and g.name == tmp[1], all_groups))
                        assert len(tmp_group) == 1
                        group = tmp_group[0]
                    else:
                        # we have ':groupname' and add to current project
                        group = new_groups.get(tmp[1])
                        assert group is not None
                    group_ids.add(group.id)
                else:
                    if tmp[0] == "SystemAdmin":
                        sysadmin = True

            project_infos: Dict[str, bool] = {}
            for projectname in user["projects"]:
                #
                # now we determine the project memberships of the user
                # projectname has the form [projectname]:"member"|"admin" (projectname omitted = current project)
                #
                tmp = projectname.split(':')
                assert len(tmp) == 2
                if tmp[0]:
                    # we have 'proj_shortname:"member"|"admin"'
                    if not all_projects:
                        all_projects = project.getAllProjects(con)
                    tmp_project = list(filter(lambda g: g.shortname == tmp[0], all_projects))
                    assert len(tmp_project) == 1
                    in_project = tmp_project[0]
                else:
                    # we have ':"member"|"admin"'
                    in_project = project
                if tmp[1] == "admin":
                    project_infos[in_project.id] = True
                else:
                    project_infos[in_project.id] = False
            user_existing = False;
            tmp_user = None
            try:
                tmp_user = User(con, username=user["username"]).read()
            except BaseError as err:
                pass
            if tmp_user is None:
                try:
                    tmp_user = User(con, email=user["email"]).read()
                except BaseError as err:
                    pass
            if tmp_user:
                #
                # The user is already in the database – let's update its settings
                #
                if tmp_user.username != user["username"]:
                    tmp_user.username = user["username"]
                if tmp_user.email != user["email"]:
                    tmp_user.email = user["email"]
                if tmp_user.givenName != user["givenName"]:
                    tmp_user.givenName = user["givenName"]
                if tmp_user.familyName != user["familyName"]:
                    tmp_user.familyName = user["familyName"]
                if tmp_user.password != user["password"]:
                    tmp_user.password = user["password"]
                if user.get("status") and tmp_user.status != user["status"]:
                    tmp_user.status = user["status"]
                if user.get("lang") and tmp_user.lang != user["lang"]:
                    tmp_user.lang = user["lang"]
                if tmp_user.sysadmin != sysadmin:
                    tmp_user.sysadmin = sysadmin
                try:
                    tmp_user.update()
                except BaseError as err:
                    tmp_user.print()
                    print("Updating user failed: " + err.message)
                    return False
                #
                # now we update group and project membership
                # Note: we do NOT remove any mambership here, we just add!
                #
                tmp_in_groups = tmp_user.in_groups
                add_groups = group_ids - tmp_in_groups
                for g in add_groups:
                    User.addToGroup(g)
                rm_groups = tmp_in_groups - group_ids
                # we do no remove a user from a group here!
                tmp_in_projects = tmp_user.in_projects
                for p in project_infos.items():
                    if tmp_in_projects.get(p[0]) and tmp_in_projects[p[0]] == p[1]:
                        continue
                    User.addToProject(p[0], p[1])
            else:
                #
                # The user does not exist yet, let's create a new one
                #
                try:
                    new_user = User(
                        con=con,
                        username=user["username"],
                        email=user["email"],
                        givenName=user["givenName"],
                        familyName=user["familyName"],
                        password=user["password"],
                        status=user["status"] if user.get("status") is not None else True,
                        lang=user["lang"] if user.get("lang") is not None else "en",
                        sysadmin=sysadmin,
                        in_projects=project_infos,
                        in_groups=group_ids
                    ).create()
                except BaseError as err:
                    print("Creating user failed: " + err.message)
                    return False
            if verbose:
                print("New user:")
                new_user.print()

    # --------------------------------------------------------------------------
    # now let's create the ontologies
    #
    ontologies = datamodel["project"]["ontologies"]
    for ontology in ontologies:
        newontology = Ontology(
            con=con,
            project=project,
            label=ontology["label"],
            name=ontology["name"]
        ).create()
        last_modification_date = newontology.lastModificationDate
        if verbose:
            print("Created empty ontology:")
            newontology.print()

        #
        # add prefixes defined in json file...
        #
        for prefix, iri in context:
            if not prefix in newontology.context:
                s = iri.iri + ("#" if iri.hashtag else "")
                newontology.context.add_context(prefix, s)

        #
        # First we create the empty resource classes
        #
        resclasses = ontology["resources"]
        newresclasses: Dict[str, ResourceClass] = {}
        for resclass in resclasses:
            resname = resclass.get("name")
            super_classes = resclass.get("super")
            if isinstance(super_classes, str):
                super_classes = [super_classes]
            reslabel = LangString(resclass.get("labels"))
            rescomment = resclass.get("comment")
            if rescomment is not None:
                rescomment = LangString(rescomment)
            try:
                last_modification_date, newresclass = ResourceClass(
                    con=con,
                    context=newontology.context,
                    ontology_id=newontology.id,
                    name=resname,
                    superclasses=super_classes,
                    label=reslabel,
                    comment=rescomment
                ).create(last_modification_date)
                newontology.lastModificationDate = last_modification_date
            except BaseError as err:
                print("Creating resource class failed: " + err.message)
                exit(105)
            newresclasses[newresclass.id] = newresclass
            if verbose is not None:
                print("New resource class:")
                newresclass.print()

        #
        # Then we create the property classes
        #
        propclasses = ontology["properties"]
        newpropclasses: Dict[str, ResourceClass] = {}
        for propclass in propclasses:
            propname = propclass.get("name")
            proplabel = LangString(propclass.get("labels"))
            #
            # get the super-property/ies if defined. Valid forms are:
            #   - "prefix:superproperty" : fully qualified name of property in another ontology. The prefix has to
            #     be defined in the prefixes part.
            #   - "superproperty" : Use of super-property defined in the knora-api ontology
            #  if omitted, automatically "knora-api:hasValue" is assumed
            #
            if propclass.get("super") is not None:
                super_props = list(map(lambda a: a if ':' in a else "knora-api:" + a, propclass["super"]))
            else:
                super_props = ["knora-api:hasValue"]
            #
            # now we get the "object" if defined. Valid forms are:
            #  - "prefix:object_name" : fully qualified object. The prefix has to be defined in the prefixes part.
            #  - ":object_name" : The object is defined in the current ontology.
            #  - "object_name" : The object is defined in "knora-api"
            #
            if propclass.get("object") is not None:
                tmp = propclass["object"].split(':')
                if len(tmp) > 1:
                    if tmp[0]:
                        object = propclass["object"] # fully qualified name
                    else:
                        newontology.print()
                        object = newontology.name + ':' + tmp[1]
                else:
                    object = "knora-api:" + propclass["object"]
            else:
                object = None

            if propclass.get("subject") is not None:
                subject = propclass["subject"]
            else:
                subject = None
            gui_element = propclass.get("gui_element")
            gui_attributes = propclass.get("gui_attributes")
            if gui_attributes is not None and gui_attributes.get("hlist") is not None:
                gui_attributes['hlist'] = "<" + listrootnodes[gui_attributes['hlist']]["id"] + ">"
            propcomment = propclass.get("comment")
            if propcomment is not None:
                propcomment = LangString(propcomment)
            else:
                propcomment = "no comment given"
            try:
                last_modification_date, newpropclass = PropertyClass(
                    con=con,
                    context=newontology.context,
                    label=proplabel,
                    name=propname,
                    ontology_id=newontology.id,
                    superproperties=super_props,
                    object=object,
                    subject=subject,
                    gui_element="salsah-gui:" + gui_element,
                    gui_attributes=gui_attributes,
                    comment=propcomment
                ).create(last_modification_date)
                newontology.lastModificationDate = last_modification_date
            except BaseError as err:
                print("Creating property class failed: " + err.message)
                return False
            newpropclasses[newpropclass.id] = newpropclass
            if verbose:
                print("New property class:")
                newpropclass.print()

        #
        # Add cardinalities
        #
        switcher = {
            "1": Cardinality.C_1,
            "0-1": Cardinality.C_0_1,
            "0-n": Cardinality.C_0_n,
            "1-n": Cardinality.C_1_n
        }
        for resclass in resclasses:
            for cardinfo in resclass["cardinalities"]:
                rc = newresclasses.get(newontology.id + '#' + resclass["name"])
                cardinality = switcher[cardinfo["cardinality"]]
                tmp = cardinfo["propname"].split(':')
                if len(tmp) > 1:
                    if tmp[0]:
                        propid = cardinfo["propname"]  # fully qualified name
                    else:
                        propid = newontology.name + ':' + tmp[1]
                else:
                    propid = "knora-api:" + cardinfo["propname"]
                gui_order = cardinfo.get('gui_order')
                last_modification_date = rc.addProperty(property_id=propid,
                                                        cardinality=cardinality,
                                                        gui_order=gui_order,
                                                        last_modification_date=last_modification_date)
                newontology.lastModificationDate = last_modification_date
    return True



from typing import List, Set, Dict, Tuple, Optional, Any, Union

from .helpers import Actions, BaseError
from .connection import Connection

class Model:
    _con: Connection
    _changed: Set[str]

    def __init__(self, con: Connection):
        if not isinstance(con, Connection):
            raise BaseError('"con"-parameter must be an instance of Connection')
        self._con = con
        self._changed = set()

    def has_changed(self) -> bool:
        if self._changed:
            return True
        else:
            return False
